from torch import nn

from ..interface import DrawingInterface
from perceptor.transforms import resize
from .init.fractal import fractal
from .init.gradient import gradient


class Raw(DrawingInterface):
    def __init__(self, init_images):
        """
        Minimal container for a nn.Parameter with init helpers.

        Usage:

            images = Raw(Raw.random_fractal_image((1, 3, 256, 256)))
        """
        super().__init__()
        self.images = nn.Parameter(init_images)

    def synthesize(self, _=None):
        return self.images

    def encode(self, images, mode="bilinear"):
        return resize(
            images,
            out_shape=tuple(self.images.shape[-2:]),
            resample=mode,
        )

    def replace_(self, images):
        self.images.data.copy_(images.data)
        return self

    @staticmethod
    def random_fractal_image(shape):
        return fractal(shape)

    @staticmethod
    def random_gradient_image(shape):
        return gradient(shape)


def test_raw():
    import torch

    Raw(torch.zeros(1, 3, 128, 128))


def test_raw_fractal():
    Raw.random_fractal_image((1, 3, 256, 256))


def test_raw_gradient():
    Raw.random_gradient_image((1, 3, 256, 256))



import torch

from perceptor import models
from perceptor.losses.interface import LossInterface


class SLIP(LossInterface):
    def __init__(self, name="SLIP_CC12M"):
        """
        Args:
            name: name of the slip model. Available models are:
                - SLIP_VITS16
                - SLIP_VITB16
                - SLIP_VITL16
                - CLIP_VITS16
                - CLIP_VITB16
                - CLIP_VITL16
                - SLIP_CC3M
                - SLIP_CC12M
        """
        super().__init__()
        self.name = name
        self.model = models.SLIP(name)
        self.encodings = None
        self.weights = None

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def add_texts_(self, texts, weights=None):
        return self.add_encodings_(self.model.encode_texts(texts), weights)

    def add_images_(self, images, weights=None):
        return self.add_encodings_(self.model.encode_images(images), weights)

    def add_encodings_(
        self,
        encodings,
        weights=None,
    ):
        if isinstance(weights, list) or isinstance(weights, tuple):
            weights = torch.tensor(weights)
        elif weights is None:
            weights = torch.ones_like(encodings[:, 0])

        if self.encodings is None:
            self.encodings = torch.nn.Parameter(
                encodings.to(self.device), requires_grad=False
            )
            self.weights = torch.nn.Parameter(
                weights.to(self.device),
                requires_grad=False,
            )
        else:
            self.encodings = torch.nn.Parameter(
                torch.cat([self.encodings, encodings.to(self.device)]),
                requires_grad=False,
            )
            self.weights = torch.nn.Parameter(
                torch.cat([self.weights, weights.to(self.device)]),
                requires_grad=False,
            )
        return self

    def forward(self, images):
        image_encodings = self.model.encode_images(images)
        spherical_distance = (
            (image_encodings[:, None] - self.encodings[None, :])
            .norm(dim=2)
            .div(2)
            .arcsin()
            .square()
            .mul(2)
        )
        return (spherical_distance * self.weights).mean()


def test_slip_loss():
    loss = (
        SLIP().add_texts_(["hello", "world"]).add_images_(torch.randn(1, 3, 256, 256))
    )
    loss(torch.randn(1, 3, 256, 256))



import torch
from torch import nn

from .interface import DrawingInterface
from perceptor import models


class DeepImagePrior(DrawingInterface):
    def __init__(self, size, n_feature_channels=64, output_channels=3):
        super().__init__()
        self.deep_image_prior = models.DeepImagePrior(
            shape=(n_feature_channels, *size), output_channels=output_channels
        )
        self.latents = nn.Parameter(
            self.deep_image_prior.random_latents(), requires_grad=False
        )
        self.images = nn.Parameter(torch.zeros((1, output_channels, *size)))

    def synthesize(self, _=None):
        return self.deep_image_prior(self.latents) + self.images

    def loss(self):
        return self.images.abs().mean().mul(0.0001)



from torchvision import transforms
from resmem import ResMem

from .interface import LossInterface


class Memorability(LossInterface):
    def __init__(self):
        super().__init__()
        self.recenter = transforms.Compose(
            [
                transforms.Resize(256),
                transforms.CenterCrop(227),
            ]
        )
        self.model = ResMem(pretrained=True)
        self.model.eval()
        self.model.requires_grad_(False)

    def forward(self, images):
        image_x = self.recenter(images)
        prediction = self.model(image_x)
        return prediction.mean() * 0.05



from torch import nn

from perceptor.drawers.interface import DrawingInterface
from perceptor import models


class StyleGANXL(DrawingInterface):
    def __init__(self, n_images=1, name="imagenet128"):
        super().__init__()
        self.model = models.StyleGANXL(name)
        self.latents = nn.Parameter(self.model.latents(n_images), requires_grad=True)

    def synthesize(self, _=None):
        return self.decode(self.latents)

    def encode(self, images):
        raise NotImplementedError()

    def decode(self, latents):
        return self.model(latents)



from perceptor.transforms.resize import resize
from .interface import LossInterface


class Resize(LossInterface):
    def __init__(self, size=None):
        super().__init__()
        self.size = size

    def forward(self, images_a, images_b, size=None):
        if size is None:
            size = self.size

        return (
            (resize(images_a, out_shape=size) - resize(images_b, out_shape=size))
            .square()
            .mean()
        )



import torch
import torch.nn.functional as F
import torchvision.transforms
import open_clip

from perceptor import utils
from perceptor.transforms.resize import resize


@utils.cache
class OpenCLIP(torch.nn.Module):
    def __init__(
        self,
        architecture="ViT-H-14",
        weights="laion2b_s32b_b79k",
        precision=None,
        jit=False,
    ):
        """
        Args:
            architecture (str): name of the clip model
            weights (str): name of the weights

            Available weight/model combinations are (in order of relevance):
            - ("ViT-H-14", "laion2b_s32b_b79k") (78.0%)
            - ("ViT-g-14", "laion2b_s12b_b42k") (76.6%)
            - ("ViT-L-14", "laion2b_s32b_b82k") (75.3%)
            - ("ViT-B-32", "laion2b_s34b_b79k") (66.6%)
            - ("ViT-B-16-plus-240", "laion400m_e32") (69.2%)
            - ("ViT-B-32", "laion2b_e16") (65.7%)
            - ("ViT-B-16", "laion400m_e32") (67.0%)
            - ("ViT-B-32", "laion400m_e32") (62.9%)
            - ("ViT-L-14", "laion400m_e32") (72.8%)
            - ("RN101", "yfcc15m") (34.8%)
            - ("RN50", "yfcc15m") (32.7%)
            - ("RN50", "cc12m") (36.45%)
            - ("RN50-quickgelu", "openai") (59.6%)
            - ("RN101-quickgelu", "openai")
            - ("RN50x4", "openai")
            - ("RN50x16", "openai")
            - ("RN50x64", "openai")
            - ("ViT-B-32-quickgelu", "openai") (63.3%)
            - ("ViT-B-16", "openai") (68.3%)
            - ("ViT-L-14", "openai") (75.6%)
            - ("ViT-L-14-336", "openai") (76.6%)
        """
        super().__init__()
        self.architecture = architecture
        self.weights = weights

        if (architecture, weights) not in open_clip.list_pretrained():
            raise ValueError(f"Invalid architecture/weights: {architecture}/{weights}")

        # softmax on cpu does not support half precision
        start_device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        if precision is None:
            if start_device == torch.device("cuda"):
                precision = "fp16"
            else:
                precision = "fp32"

        self.model, _, transforms = open_clip.create_model_and_transforms(
            architecture,
            weights,
            device=start_device,
            precision=precision,
            jit=jit,
            cache_dir="models",
        )
        self.model = self.model.eval()

        if jit is False:
            self.model = self.model.requires_grad_(False)

        self.normalize = torchvision.transforms.Normalize(
            transforms.transforms[-1].mean,
            transforms.transforms[-1].std,
        )

    def to(self, device):
        if device == torch.device("cpu"):
            self.model = self.model.float()
        return super().to(device)

    @property
    def device(self):
        return next(iter(self.parameters())).device

    @property
    def image_size(self):
        if isinstance(self.model.visual.image_size, tuple):
            return self.model.visual.image_size
        else:
            return (self.model.visual.image_size, self.model.visual.image_size)

    @torch.cuda.amp.autocast()
    def encode_texts(self, text_prompts, normalize=True):
        encodings = self.model.encode_text(
            open_clip.tokenize(text_prompts).to(self.device)
        )
        if normalize:
            return F.normalize(encodings)
        else:
            return encodings

    @torch.cuda.amp.autocast()
    def encode_images(self, images, normalize=True):
        encodings = self.model.encode_image(
            self.normalize(
                resize(
                    images.to(self.device),
                    out_shape=self.image_size,
                )
            )
        )

        if normalize:
            return F.normalize(encodings)
        else:
            return encodings

    def forward(self, _):
        raise NotImplementedError


def test_open_clip():
    import torch

    model = OpenCLIP("ViT-B-32", "laion2b_s34b_b79k")

    image = torch.randn((1, 3, 256, 256)).requires_grad_()
    with torch.enable_grad():
        model.encode_images(image).mean().backward()

    assert image.grad is not None



import math
import torch
from torch.nn import functional as F


class RealESRGANer(torch.nn.Module):
    """A helper class for upsampling images with RealESRGAN.
    Args:
        scale (int): Upsampling scale factor used in the networks. It is usually 2 or 4.
        model_path (str): The path to the pretrained model
        model (nn.Module): The defined network. Default: None.
        tile (int): As too large images result in the out of GPU memory issue, so this tile option will first crop
            input images into tiles, and then process each of them. Finally, they will be merged into one image.
            0 denotes for do not use tile. Default: 0.
        tile_pad (int): The pad size for each tile, to remove border artifacts. Default: 10.
        pre_pad (int): Pad the input images to avoid border artifacts. Default: 10.
        half (float): Whether to use half precision during inference. Default: False.
    """

    def __init__(
        self, scale, model_path, model, tile=0, tile_pad=10, pre_pad=10, half=False
    ):
        super().__init__()
        self.scale = scale
        self.model = model
        self.tile_size = tile
        self.tile_pad = tile_pad
        self.pre_pad = pre_pad
        self.mod_scale = None
        self.half = half

        # initialize model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # if the model_path starts with https, it will first download models to the folder: realesrgan/weights

        loadnet = torch.load(model_path, map_location=torch.device("cpu"))
        # prefer to use params_ema
        if "params_ema" in loadnet:
            self.model.load_state_dict(loadnet["params_ema"], strict=True)
        elif "params" in loadnet:
            self.model.load_state_dict(loadnet["params"], strict=True)
        else:
            self.model.load_state_dict(loadnet, strict=True)

        self.model.eval()
        self.model = model.to(self.device)
        if self.half:
            self.model = self.model.half()

    def pre_process(self, img):
        """Pre-process, such as pre-pad and mod pad, so that the images can be divisible"""
        # img is pytorch tensor
        self.img = img.clone().to(self.device)
        if self.half:
            self.img = self.img.half()

        # pre_pad
        if self.pre_pad != 0:
            self.img = F.pad(self.img, (0, self.pre_pad, 0, self.pre_pad), "reflect")
        # mod pad for divisible borders
        if self.scale == 2:
            self.mod_scale = 2
        elif self.scale == 1:
            self.mod_scale = 4
        if self.mod_scale is not None:
            self.mod_pad_h, self.mod_pad_w = 0, 0
            _, _, h, w = self.img.size()
            if h % self.mod_scale != 0:
                self.mod_pad_h = self.mod_scale - h % self.mod_scale
            if w % self.mod_scale != 0:
                self.mod_pad_w = self.mod_scale - w % self.mod_scale
            self.img = F.pad(
                self.img, (0, self.mod_pad_w, 0, self.mod_pad_h), "reflect"
            )

    def process(self):
        # model inference
        self.output = self.model(self.img)

    def tile_process(self):
        """It will first crop input images to tiles, and then process each tile.
        Finally, all the processed tiles are merged into one images.
        Modified from: https://github.com/ata4/esrgan-launcher
        """
        batch, channel, height, width = self.img.shape
        output_height = height * self.scale
        output_width = width * self.scale
        output_shape = (batch, channel, output_height, output_width)

        # start with black image
        self.output = self.img.new_zeros(output_shape)
        tiles_x = math.ceil(width / self.tile_size)
        tiles_y = math.ceil(height / self.tile_size)

        # loop over all tiles
        for y in range(tiles_y):
            for x in range(tiles_x):
                # extract tile from input image
                ofs_x = x * self.tile_size
                ofs_y = y * self.tile_size
                # input tile area on total image
                input_start_x = ofs_x
                input_end_x = min(ofs_x + self.tile_size, width)
                input_start_y = ofs_y
                input_end_y = min(ofs_y + self.tile_size, height)

                # input tile area on total image with padding
                input_start_x_pad = max(input_start_x - self.tile_pad, 0)
                input_end_x_pad = min(input_end_x + self.tile_pad, width)
                input_start_y_pad = max(input_start_y - self.tile_pad, 0)
                input_end_y_pad = min(input_end_y + self.tile_pad, height)

                # input tile dimensions
                input_tile_width = input_end_x - input_start_x
                input_tile_height = input_end_y - input_start_y
                tile_idx = y * tiles_x + x + 1
                input_tile = self.img[
                    :,
                    :,
                    input_start_y_pad:input_end_y_pad,
                    input_start_x_pad:input_end_x_pad,
                ]

                # upscale tile
                try:
                    # with torch.no_grad():
                    output_tile = self.model(input_tile)
                except RuntimeError as error:
                    print("Error", error)
                print(f"\tTile {tile_idx}/{tiles_x * tiles_y}")

                # output tile area on total image
                output_start_x = input_start_x * self.scale
                output_end_x = input_end_x * self.scale
                output_start_y = input_start_y * self.scale
                output_end_y = input_end_y * self.scale

                # output tile area without padding
                output_start_x_tile = (input_start_x - input_start_x_pad) * self.scale
                output_end_x_tile = output_start_x_tile + input_tile_width * self.scale
                output_start_y_tile = (input_start_y - input_start_y_pad) * self.scale
                output_end_y_tile = output_start_y_tile + input_tile_height * self.scale

                # put tile into output image
                self.output[
                    :, :, output_start_y:output_end_y, output_start_x:output_end_x
                ] = output_tile[
                    :,
                    :,
                    output_start_y_tile:output_end_y_tile,
                    output_start_x_tile:output_end_x_tile,
                ]

    def post_process(self):
        # remove extra pad
        if self.mod_scale is not None:
            _, _, h, w = self.output.size()
            self.output = self.output[
                :,
                :,
                0 : h - self.mod_pad_h * self.scale,
                0 : w - self.mod_pad_w * self.scale,
            ]
        # remove prepad
        if self.pre_pad != 0:
            _, _, h, w = self.output.size()
            self.output = self.output[
                :,
                :,
                0 : h - self.pre_pad * self.scale,
                0 : w - self.pre_pad * self.scale,
            ]
        return self.output

    # @torch.no_grad()
    def enhance(self, img, outscale=None, alpha_upsampler="realesrgan"):
        h_input, w_input = img.shape[-2:]

        self.pre_process(img)
        if self.tile_size > 0:
            self.tile_process()
        else:
            self.process()
        output_img = self.post_process()
        # output_img = output_img.data.squeeze().float().cpu().clamp_(0, 1).numpy()
        return output_img



import torch
from torch import nn

from perceptor.drawers.interface import DrawingInterface


class BruteDiffusion(DrawingInterface):
    def __init__(self, model, diffused_images, t):
        super().__init__()
        self.model = model
        self.t = t
        self.diffused_images = nn.Parameter(
            diffused_images,
            requires_grad=True,
        )

    @staticmethod
    def from_image(model, images, t, noise=None):
        drawer = BruteDiffusion(
            model,
            torch.zeros_like(images),
            t,
        )
        return drawer.replace_(drawer.encode(images, noise))

    @property
    def x(self):
        return self.diffused_images.mul(2).sub(1)

    def synthesize(self, _=None):
        return (
            self.model.predict_denoised(self.diffused_images.mul(2).sub(1), self.t)
            .add(1)
            .div(2)
        )

    def encode(self, images, noise=None):
        return (
            self.model.diffuse(images.mul(2).sub(1), t=self.t, noise=noise)
            .add(1)
            .div(2)
        )

    def replace_(self, diffused_images):
        self.diffused_images.data.copy_(diffused_images)
        return self

    def noise(self):
        return self.model.noise(self.x, self.t)



from .interface import LossInterface


class Smoothness(LossInterface):
    def forward(self, images):
        images = images.contiguous().clone()
        gradient_height = images[:, :, 1:, :] - images[:, :, :-1, :]
        gradient_width = images[:, :, :, 1:] - images[:, :, :, :-1]
        sharpness = gradient_height.square().mean() + gradient_width.square().mean()
        return sharpness



import torch.nn as nn
import torch
import torch.nn as nn
import torch.nn.functional as F
from .common import *


class ListModule(nn.Module):
    def __init__(self, *args):
        super(ListModule, self).__init__()
        idx = 0
        for module in args:
            self.add_module(str(idx), module)
            idx += 1

    def __getitem__(self, idx):
        if idx >= len(self._modules):
            raise IndexError("index {} is out of range".format(idx))
        if idx < 0:
            idx = len(self) + idx

        it = iter(self._modules.values())
        for i in range(idx):
            next(it)
        return next(it)

    def __iter__(self):
        return iter(self._modules.values())

    def __len__(self):
        return len(self._modules)


class UNet(nn.Module):
    """
    upsample_mode in ['deconv', 'nearest', 'bilinear']
    pad in ['zero', 'replication', 'none']
    """

    def __init__(
        self,
        num_input_channels=3,
        num_output_channels=3,
        feature_scale=4,
        more_layers=0,
        concat_x=False,
        upsample_mode="deconv",
        pad="zero",
        norm_layer=nn.InstanceNorm2d,
        need_sigmoid=True,
        need_bias=True,
    ):
        super(UNet, self).__init__()

        self.feature_scale = feature_scale
        self.more_layers = more_layers
        self.concat_x = concat_x

        filters = [64, 128, 256, 512, 1024]
        filters = [x // self.feature_scale for x in filters]

        self.start = unetConv2(
            num_input_channels,
            filters[0] if not concat_x else filters[0] - num_input_channels,
            norm_layer,
            need_bias,
            pad,
        )

        self.down1 = unetDown(
            filters[0],
            filters[1] if not concat_x else filters[1] - num_input_channels,
            norm_layer,
            need_bias,
            pad,
        )
        self.down2 = unetDown(
            filters[1],
            filters[2] if not concat_x else filters[2] - num_input_channels,
            norm_layer,
            need_bias,
            pad,
        )
        self.down3 = unetDown(
            filters[2],
            filters[3] if not concat_x else filters[3] - num_input_channels,
            norm_layer,
            need_bias,
            pad,
        )
        self.down4 = unetDown(
            filters[3],
            filters[4] if not concat_x else filters[4] - num_input_channels,
            norm_layer,
            need_bias,
            pad,
        )

        # more downsampling layers
        if self.more_layers > 0:
            self.more_downs = [
                unetDown(
                    filters[4],
                    filters[4] if not concat_x else filters[4] - num_input_channels,
                    norm_layer,
                    need_bias,
                    pad,
                )
                for i in range(self.more_layers)
            ]
            self.more_ups = [
                unetUp(filters[4], upsample_mode, need_bias, pad, same_num_filt=True)
                for i in range(self.more_layers)
            ]

            self.more_downs = ListModule(*self.more_downs)
            self.more_ups = ListModule(*self.more_ups)

        self.up4 = unetUp(filters[3], upsample_mode, need_bias, pad)
        self.up3 = unetUp(filters[2], upsample_mode, need_bias, pad)
        self.up2 = unetUp(filters[1], upsample_mode, need_bias, pad)
        self.up1 = unetUp(filters[0], upsample_mode, need_bias, pad)

        self.final = conv(filters[0], num_output_channels, 1, bias=need_bias, pad=pad)

        if need_sigmoid:
            self.final = nn.Sequential(self.final, nn.Sigmoid())

    def forward(self, inputs):

        # Downsample
        downs = [inputs]
        down = nn.AvgPool2d(2, 2)
        for i in range(4 + self.more_layers):
            downs.append(down(downs[-1]))

        in64 = self.start(inputs)
        if self.concat_x:
            in64 = torch.cat([in64, downs[0]], 1)

        down1 = self.down1(in64)
        if self.concat_x:
            down1 = torch.cat([down1, downs[1]], 1)

        down2 = self.down2(down1)
        if self.concat_x:
            down2 = torch.cat([down2, downs[2]], 1)

        down3 = self.down3(down2)
        if self.concat_x:
            down3 = torch.cat([down3, downs[3]], 1)

        down4 = self.down4(down3)
        if self.concat_x:
            down4 = torch.cat([down4, downs[4]], 1)

        if self.more_layers > 0:
            prevs = [down4]
            for kk, d in enumerate(self.more_downs):
                # print(prevs[-1].size())
                out = d(prevs[-1])
                if self.concat_x:
                    out = torch.cat([out, downs[kk + 5]], 1)

                prevs.append(out)

            up_ = self.more_ups[-1](prevs[-1], prevs[-2])
            for idx in range(self.more_layers - 1):
                l = self.more_ups[self.more - idx - 2]
                up_ = l(up_, prevs[self.more - idx - 2])
        else:
            up_ = down4

        up4 = self.up4(up_, down3)
        up3 = self.up3(up4, down2)
        up2 = self.up2(up3, down1)
        up1 = self.up1(up2, in64)

        return self.final(up1)


class unetConv2(nn.Module):
    def __init__(self, in_size, out_size, norm_layer, need_bias, pad):
        super(unetConv2, self).__init__()

        print(pad)
        if norm_layer is not None:
            self.conv1 = nn.Sequential(
                conv(in_size, out_size, 3, bias=need_bias, pad=pad),
                norm_layer(out_size),
                nn.ReLU(),
            )
            self.conv2 = nn.Sequential(
                conv(out_size, out_size, 3, bias=need_bias, pad=pad),
                norm_layer(out_size),
                nn.ReLU(),
            )
        else:
            self.conv1 = nn.Sequential(
                conv(in_size, out_size, 3, bias=need_bias, pad=pad),
                nn.ReLU(),
            )
            self.conv2 = nn.Sequential(
                conv(out_size, out_size, 3, bias=need_bias, pad=pad),
                nn.ReLU(),
            )

    def forward(self, inputs):
        outputs = self.conv1(inputs)
        outputs = self.conv2(outputs)
        return outputs


class unetDown(nn.Module):
    def __init__(self, in_size, out_size, norm_layer, need_bias, pad):
        super(unetDown, self).__init__()
        self.conv = unetConv2(in_size, out_size, norm_layer, need_bias, pad)
        self.down = nn.MaxPool2d(2, 2)

    def forward(self, inputs):
        outputs = self.down(inputs)
        outputs = self.conv(outputs)
        return outputs


class unetUp(nn.Module):
    def __init__(self, out_size, upsample_mode, need_bias, pad, same_num_filt=False):
        super(unetUp, self).__init__()

        num_filt = out_size if same_num_filt else out_size * 2
        if upsample_mode == "deconv":
            self.up = nn.ConvTranspose2d(num_filt, out_size, 4, stride=2, padding=1)
            self.conv = unetConv2(out_size * 2, out_size, None, need_bias, pad)
        elif upsample_mode == "bilinear" or upsample_mode == "nearest":
            self.up = nn.Sequential(
                nn.Upsample(scale_factor=2, mode=upsample_mode),
                conv(num_filt, out_size, 3, bias=need_bias, pad=pad),
            )
            self.conv = unetConv2(out_size * 2, out_size, None, need_bias, pad)
        else:
            assert False

    def forward(self, inputs1, inputs2):
        in1_up = self.up(inputs1)

        if (inputs2.size(2) != in1_up.size(2)) or (inputs2.size(3) != in1_up.size(3)):
            diff2 = (inputs2.size(2) - in1_up.size(2)) // 2
            diff3 = (inputs2.size(3) - in1_up.size(3)) // 2
            inputs2_ = inputs2[
                :, :, diff2 : diff2 + in1_up.size(2), diff3 : diff3 + in1_up.size(3)
            ]
        else:
            inputs2_ = inputs2

        output = self.conv(torch.cat([in1_up, inputs2_], 1))

        return output



import torch
import torch.nn.functional as F

from perceptor.losses.interface import LossInterface
from perceptor import transforms


class SuperResolution(LossInterface):
    def __init__(self, name="x2", pre_downscale=None, half=True, mode="bicubic"):
        super().__init__()
        self.transform = transforms.SuperResolution(name, half)
        self.mode = mode
        self.pre_downscale = (
            self.transform.model.scale if pre_downscale is None else pre_downscale
        )

    def forward(self, images):
        with torch.no_grad():
            downsampled_size = torch.tensor(images.shape[-2:]) // self.pre_downscale
            scale = downsampled_size / torch.tensor(images.shape[-2:])
            downsampled = transforms.resize(
                images, scale.tolist(), downsampled_size, resample=self.mode
            )
            upsampled = self.transform(downsampled)
            if upsampled.shape != images.shape:
                scale = torch.tensor(images.shape[-2:]) / torch.tensor(
                    upsampled.shape[-2:]
                )
                upsampled = transforms.resize(
                    upsampled,
                    scale.tolist(),
                    out_shape=tuple(images.shape[-2:]),
                    resample=self.mode,
                )
        return (images - upsampled).square().mean()



import os
import re
import json

from typing import List, Tuple, Union, Optional
from collections import OrderedDict
from locale import atof

import click
import numpy as np
import torch
import torch.nn.functional as F

from .. import dnnlib


# ----------------------------------------------------------------------------


def create_image_grid(images: np.ndarray, grid_size: Optional[Tuple[int, int]] = None):
    """
    Create a grid with the fed images
    Args:
        images (np.array): array of images
        grid_size (tuple(int)): size of grid (grid_width, grid_height)
    Returns:
        grid (np.array): image grid of size grid_size
    """
    # Sanity check
    assert (
        images.ndim == 3 or images.ndim == 4
    ), f"Images has {images.ndim} dimensions (shape: {images.shape})!"
    num, img_h, img_w, c = images.shape
    # If user specifies the grid shape, use it
    if grid_size is not None:
        grid_w, grid_h = tuple(grid_size)
        # If one of the sides is None, then we must infer it (this was divine inspiration)
        if grid_w is None:
            grid_w = num // grid_h + min(num % grid_h, 1)
        elif grid_h is None:
            grid_h = num // grid_w + min(num % grid_w, 1)

    # Otherwise, we can infer it by the number of images (priority is given to grid_w)
    else:
        grid_w = max(int(np.ceil(np.sqrt(num))), 1)
        grid_h = max((num - 1) // grid_w + 1, 1)

    # Sanity check
    assert (
        grid_w * grid_h >= num
    ), "Number of rows and columns must be greater than the number of images!"
    # Get the grid
    grid = np.zeros(
        [grid_h * img_h, grid_w * img_h] + list(images.shape[-1:]), dtype=images.dtype
    )
    # Paste each image in the grid
    for idx in range(num):
        x = (idx % grid_w) * img_w
        y = (idx // grid_w) * img_h
        grid[y : y + img_h, x : x + img_w, ...] = images[idx]
    return grid


# ----------------------------------------------------------------------------


def parse_fps(fps: Union[str, int]) -> int:
    """Return FPS for the video; at worst, video will be 1 FPS, but no lower.
    Useful if we don't have Click, else simply use Click.IntRange(min=1)"""
    if isinstance(fps, int):
        return max(fps, 1)
    try:
        fps = int(atof(fps))
        return max(fps, 1)
    except ValueError:
        print(f'Typo in "--fps={fps}", will use default value of 30')
        return 30


def num_range(s: str, remove_repeated: bool = True) -> List[int]:
    """
    Extended helper function from the original (original is contained here).
    Accept a comma separated list of numbers 'a,b,c', a range 'a-c', or a combination
    of both 'a,b-c', 'a-b,c', 'a,b-c,d,e-f,...', and return as a list of ints.
    """
    str_list = s.split(",")
    nums = []
    range_re = re.compile(r"^(\d+)-(\d+)$")
    for el in str_list:
        match = range_re.match(el)
        if match:
            # Sanity check 1: accept ranges 'a-b' or 'b-a', with a<=b
            lower, upper = int(match.group(1)), int(match.group(2))
            if lower <= upper:
                r = list(range(lower, upper + 1))
            else:
                r = list(range(upper, lower + 1))
            # We will extend nums as r is also a list
            nums.extend(r)
        else:
            # It's a single number, so just append it (if it's an int)
            try:
                nums.append(int(atof(el)))
            except ValueError:
                continue  # we ignore bad values
    # Sanity check 2: delete repeating numbers by default, but keep order given by user
    if remove_repeated:
        nums = list(OrderedDict.fromkeys(nums))
    return nums


def parse_slowdown(slowdown: Union[str, int]) -> int:
    """Function to parse the 'slowdown' parameter by the user. Will approximate to the nearest power of 2."""
    # TODO: slowdown should be any int
    if not isinstance(slowdown, int):
        try:
            slowdown = atof(slowdown)
        except ValueError:
            print(f'Typo in "{slowdown}"; will use default value of 1')
            slowdown = 1
    assert slowdown > 0, '"slowdown" cannot be negative or 0!'
    # Let's approximate slowdown to the closest power of 2 (nothing happens if it's already a power of 2)
    slowdown = 2 ** int(np.rint(np.log2(slowdown)))
    return max(slowdown, 1)  # Guard against 0.5, 0.25, ... cases


def parse_new_center(s: str) -> Tuple[str, Union[int, np.ndarray]]:
    """Get a new center for the W latent space (a seed or projected dlatent; to be transformed later)"""
    try:
        new_center = int(s)  # it's a seed
        return s, new_center
    except ValueError:
        new_center = get_w_from_file(s, return_ext=False)  # it's a projected dlatent
        return s, new_center


# ----------------------------------------------------------------------------


def compress_video(
    original_video: Union[str, os.PathLike],
    original_video_name: Union[str, os.PathLike],
    outdir: Union[str, os.PathLike],
    ctx: click.Context,
) -> None:
    """Helper function to compress the original_video using ffmpeg-python. moviepy creates huge videos, so use
    ffmpeg to 'compress' it (won't be perfect, 'compression' will depend on the video dimensions). ffmpeg
    can also be used to e.g. resize the video, make a GIF, save all frames in the video to the outdir, etc.
    """
    try:
        import ffmpeg
    except (ModuleNotFoundError, ImportError):
        ctx.fail('Missing ffmpeg! Install it via "pip install ffmpeg-python"')

    print("Compressing the video...")
    resized_video_name = os.path.join(outdir, f"{original_video_name}-compressed.mp4")
    ffmpeg.input(original_video).output(resized_video_name).run(
        capture_stdout=True, capture_stderr=True
    )
    print("Success!")


# ----------------------------------------------------------------------------


def interpolation_checks(
    t: Union[float, np.ndarray], v0: np.ndarray, v1: np.ndarray
) -> Tuple[Union[float, np.ndarray], np.ndarray, np.ndarray]:
    """Tests for the interpolation functions"""
    # Make sure 0.0<=t<=1.0
    assert np.min(t) >= 0.0 and np.max(t) <= 1.0
    # Guard against v0 and v1 not being NumPy arrays
    if not isinstance(v0, np.ndarray):
        v0 = np.array(v0)
    if not isinstance(v1, np.ndarray):
        v1 = np.array(v1)
    # Both should have the same shape in order to interpolate between them
    assert v0.shape == v1.shape, f"Incompatible shapes! v0: {v0.shape}, v1: {v1.shape}"
    return t, v0, v1


def lerp(
    t: Union[float, np.ndarray],
    v0: Union[float, list, tuple, np.ndarray],
    v1: Union[float, list, tuple, np.ndarray],
) -> np.ndarray:
    """
    Linear interpolation between v0 (starting) and v1 (final) vectors; for optimal results,
    use t as an np.ndarray to return all results at once via broadcasting
    """
    t, v0, v1 = interpolation_checks(t, v0, v1)
    v2 = (1.0 - t) * v0 + t * v1
    return v2


def slerp(
    t: Union[float, np.ndarray],
    v0: Union[float, list, tuple, np.ndarray],
    v1: Union[float, list, tuple, np.ndarray],
    dot_threshold: float = 0.9995,
) -> np.ndarray:
    """
    Spherical linear interpolation between v0 (starting) and v1 (final) vectors; for optimal
    results, use t as an np.ndarray to return all results at once via broadcasting.

    dot_threshold is the threshold for considering if the two vectors are collinear (not recommended to alter).

    Adapted from the Python code at: https://en.wikipedia.org/wiki/Slerp (at the time, now no longer available).
    Most likely taken from Jonathan Blow's code in C++:
            http://number-none.com/product/Understanding%20Slerp,%20Then%20Not%20Using%20It
    """
    t, v0, v1 = interpolation_checks(t, v0, v1)
    # Copy vectors to reuse them later
    v0_copy = np.copy(v0)
    v1_copy = np.copy(v1)
    # Normalize the vectors to get the directions and angles
    v0 = v0 / np.linalg.norm(v0)
    v1 = v1 / np.linalg.norm(v1)
    # Dot product with the normalized vectors (can't always use np.dot, so we use the definition)
    dot = np.sum(v0 * v1)
    # If it's ~1, vectors are ~colineal, so use lerp
    if np.abs(dot) > dot_threshold:
        return lerp(t, v0, v1)
    # Stay within domain of arccos
    dot = np.clip(dot, -1.0, 1.0)
    # Calculate initial angle between v0 and v1
    theta_0 = np.arccos(dot)
    sin_theta_0 = np.sin(theta_0)
    # Divide the angle into t steps
    theta_t = theta_0 * t
    sin_theta_t = np.sin(theta_t)
    # Finish the slerp algorithm
    s0 = np.sin(theta_0 - theta_t) / sin_theta_0
    s1 = sin_theta_t / sin_theta_0
    v2 = s0 * v0_copy + s1 * v1_copy
    return v2


def interpolate(
    v0: Union[float, list, tuple, np.ndarray],
    v1: Union[float, list, tuple, np.ndarray],
    n_steps: int,
    interp_type: str = "spherical",
    smooth: bool = False,
) -> np.ndarray:
    """
    Interpolation function between two vectors, v0 and v1. We will either do a 'linear' or 'spherical' interpolation,
    taking n_steps. The steps can be 'smooth'-ed out, so that the transition between vectors isn't too drastic.
    """
    t_array = np.linspace(0, 1, num=n_steps, endpoint=False)
    # TODO: have a dictionary with easing functions that contains my 'smooth' one (might be useful for someone else)
    if smooth:
        # Smooth out the interpolation with a polynomial of order 3 (cubic function f)
        # Constructed f by setting f'(0) = f'(1) = 0, and f(0) = 0, f(1) = 1 => f(t) = -2t^3+3t^2 = t^2 (3-2t)
        # NOTE: I've merely rediscovered the Smoothstep function S_1(x): https://en.wikipedia.org/wiki/Smoothstep
        t_array = t_array**2 * (3 - 2 * t_array)  # One line thanks to NumPy arrays
    # TODO: this might be possible to optimize by using the fact they're numpy arrays, but haven't found a nice way yet
    funcs_dict = {"linear": lerp, "spherical": slerp}
    vectors = np.array(
        [funcs_dict[interp_type](t, v0, v1) for t in t_array], dtype=np.float32
    )
    return vectors


# ----------------------------------------------------------------------------


def double_slowdown(
    latents: np.ndarray, duration: float, frames: int
) -> Tuple[np.ndarray, float, int]:
    """
    Auxiliary function to slow down the video by 2x. We return the new latents, duration, and frames of the video
    """
    # Make an empty latent vector with double the amount of frames, but keep the others the same
    z = np.empty(np.multiply(latents.shape, [2, 1, 1]), dtype=np.float32)
    # In the even frames, populate it with the latents
    for i in range(len(latents)):
        z[2 * i] = latents[i]
    # Interpolate in the odd frames
    for i in range(1, len(z), 2):
        # slerp between (t=0.5) even frames; for the last frame, we loop to the first one (z[0])
        z[i] = (
            slerp(0.5, z[i - 1], z[i + 1])
            if i != len(z) - 1
            else slerp(0.5, z[0], z[i - 1])
        )
    # TODO: we could change this to any slowdown: slerp(1/slowdown, ...), and we return z, slowdown * duration, ...
    # Return the new latents, and the respective new duration and number of frames
    return z, 2 * duration, 2 * frames


# ----------------------------------------------------------------------------


def make_affine_transform(
    m: torch.Tensor = None,
    angle: float = 0.0,
    translate_x: float = 0.0,
    translate_y: float = 0.0,
    scale_x: float = 1.0,
    scale_y: float = 1.0,
    shear_x: float = 0.0,
    shear_y: float = 0.0,
    mirror_x: bool = False,
    mirror_y: bool = False,
) -> np.array:
    """Make affine transformation with the given parameters. If none are passed, will return the identity.
    As a guide for affine transformations: https://en.wikipedia.org/wiki/Affine_transformation"""
    # m is the starting affine transformation matrix (e.g., G.synthesis.input.transform)
    if m is None:
        m = np.eye(3, dtype=np.float64)
    else:
        m = m.cpu().numpy()
    # Remember these are the inverse transformations!
    # Rotation matrix
    rotation_matrix = np.array(
        [
            [np.cos(angle), np.sin(angle), 0.0],
            [-np.sin(angle), np.cos(angle), 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=np.float64,
    )
    # Translation matrix
    translation_matrix = np.array(
        [[1.0, 0.0, -translate_x], [0.0, 1.0, -translate_y], [0.0, 0.0, 1.0]],
        dtype=np.float64,
    )
    # Scale matrix (don't let it go into negative or 0)
    scale_matrix = np.array(
        [
            [1.0 / max(scale_x, 1e-4), 0.0, 0.0],
            [0.0, 1.0 / max(scale_y, 1e-4), 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=np.float64,
    )
    # Shear matrix
    shear_matrix = np.array(
        [[1.0, -shear_x, 0.0], [-shear_y, 1.0, 0.0], [0.0, 0.0, 1.0]], dtype=np.float64
    )
    # Mirror/reflection in x matrix
    xmirror_matrix = np.array(
        [[1.0 - 2 * mirror_x, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
        dtype=np.float64,
    )
    # Mirror/reflection in y matrix
    ymirror_matrix = np.array(
        [[1.0, 0.0, 0.0], [0.0, 1.0 - 2 * mirror_y, 0.0], [0.0, 0.0, 1.0]],
        dtype=np.float64,
    )

    # Make the resulting affine transformation (note that these are non-commutative, so we *choose* this order)
    m = (
        m
        @ rotation_matrix
        @ translation_matrix
        @ scale_matrix
        @ shear_matrix
        @ xmirror_matrix
        @ ymirror_matrix
    )
    return m


def anchor_latent_space(G) -> None:
    # Thanks to @RiversHaveWings and @nshepperd1
    if hasattr(G.synthesis, "input"):
        shift = G.synthesis.input.affine(G.mapping.w_avg.unsqueeze(0))
        G.synthesis.input.affine.bias.data.add_(shift.squeeze(0))
        G.synthesis.input.affine.weight.data.zero_()


def force_fp32(G) -> None:
    """Force fp32 as in during training"""
    G.synthesis.num_fp16_res = 0
    for name, layer in G.synthesis.named_modules():
        if hasattr(layer, "conv_clamp"):
            layer.conv_clamp = None
            layer.use_fp16 = False


# ----------------------------------------------------------------------------

resume_specs = {
    # For StyleGAN2/ADA models; --cfg=stylegan2
    "stylegan2": {
        # Official NVIDIA models
        "ffhq256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhq-256x256.pkl",
        "ffhqu256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhqu-256x256.pkl",
        "ffhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhq-512x512.pkl",
        "ffhq1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhq-1024x1024.pkl",
        "ffhqu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-ffhqu-1024x1024.pkl",
        "celebahq256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-celebahq-256x256.pkl",
        "lsundog256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-lsundog-256x256.pkl",
        "afhqcat512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqcat-512x512.pkl",
        "afhqdog512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqdog-512x512.pkl",
        "afhqwild512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqwild-512x512.pkl",
        "afhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-afhqv2-512x512.pkl",
        "brecahad512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-brecahad-512x512.pkl",
        "cifar10": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-cifar10-32x32.pkl",
        "metfaces1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-metfaces-1024x1024.pkl",
        "metfacesu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan2/versions/1/files/stylegan2-metfacesu-1024x1024.pkl",
        # Community models; TODO: add the interesting ones found in https://github.com/justinpinkney/awesome-pretrained-stylegan2
        "minecraft1024": "https://github.com/jeffheaton/pretrained-gan-minecraft/releases/download/v1/minecraft-gan-2020-12-22.pkl",  # Thanks to @jeffheaton
        # Deceive-D/APA models (ignoring the faces models): https://github.com/EndlessSora/DeceiveD
        "afhqcat256": "https://drive.google.com/u/0/uc?export=download&confirm=zFoN&id=1P9ouHIK-W8JTb6bvecfBe4c_3w6gmMJK",
        "anime256": "https://drive.google.com/u/0/uc?export=download&confirm=6Uie&id=1EWOdieqELYmd2xRxUR4gnx7G10YI5dyP",
        "cub256": "https://drive.google.com/u/0/uc?export=download&confirm=KwZS&id=1J0qactT55ofAvzddDE_xnJEY8s3vbo1_",
    },
    # For StyleGAN3 config-r models (--cfg=stylegan3-r)
    "stylegan3-r": {
        # Official NVIDIA models
        "afhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-afhqv2-512x512.pkl",
        "ffhq1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-ffhq-1024x1024.pkl",
        "ffhqu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-ffhqu-1024x1024.pkl",
        "ffhqu256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-ffhqu-256x256.pkl",
        "metfaces1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-metfaces-1024x1024.pkl",
        "metfacesu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-r-metfacesu-1024x1024.pkl",
    },
    # For StyleGAN3 config-t models (--cfg=stylegan3-t)
    "stylegan3-t": {
        # Official NVIDIA models
        "afhq512": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-afhqv2-512x512.pkl",
        "ffhq1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-ffhq-1024x1024.pkl",
        "ffhqu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-ffhqu-1024x1024.pkl",
        "ffhqu256": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-ffhqu-256x256.pkl",
        "metfaces1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-metfaces-1024x1024.pkl",
        "metfacesu1024": "https://api.ngc.nvidia.com/v2/models/nvidia/research/stylegan3/versions/1/files/stylegan3-t-metfacesu-1024x1024.pkl",
        # Community models, found in: https://github.com/justinpinkney/awesome-pretrained-stylegan3 by @justinpinkney
        "landscapes256": "https://drive.google.com/u/0/uc?export=download&confirm=eJHe&id=14UGDDOusZ9TMb-pOrF0PAjMGVWLSAii1",  # Thanks to @justinpinkney
        "wikiart1024": "https://drive.google.com/u/0/uc?export=download&confirm=2tz5&id=18MOpwTMJsl_Z17q-wQVnaRLCUFZYSNkj",  # Thanks to @justinpinkney
    },
}

# ----------------------------------------------------------------------------


def z_to_img(
    G,
    latents: torch.Tensor,
    label: torch.Tensor,
    truncation_psi: float,
    noise_mode: str = "const",
) -> np.ndarray:
    """
    Get an image/np.ndarray from a latent Z using G, the label, truncation_psi, and noise_mode. The shape
    of the output image/np.ndarray will be [len(dlatents), G.img_resolution, G.img_resolution, G.img_channels]
    """
    assert isinstance(
        latents, torch.Tensor
    ), f'latents should be a torch.Tensor!: "{type(latents)}"'
    if len(latents.shape) == 1:
        latents = latents.unsqueeze(0)  # An individual latent => [1, G.z_dim]
    img = G(z=latents, c=label, truncation_psi=truncation_psi, noise_mode=noise_mode)
    img = (img + 1) * 255 / 2  # [-1.0, 1.0] -> [0.0, 255.0]
    img = (
        img.permute(0, 2, 3, 1).clamp(0, 255).to(torch.uint8).cpu().numpy()
    )  # NCWH => NWHC
    return img


def w_to_img(
    G,
    dlatents: Union[List[torch.Tensor], torch.Tensor],
    noise_mode: str = "const",
    to_np: bool = True,
) -> np.ndarray:
    """
    Get an image/np.ndarray from a dlatent W using G and the selected noise_mode. The final shape of the
    returned image will be [len(dlatents), G.img_resolution, G.img_resolution, G.img_channels].
    """
    assert isinstance(
        dlatents, torch.Tensor
    ), f'dlatents should be a torch.Tensor!: "{type(dlatents)}"'
    if len(dlatents.shape) == 2:
        dlatents = dlatents.unsqueeze(
            0
        )  # An individual dlatent => [1, G.mapping.num_ws, G.mapping.w_dim]

    synth_image = G.synthesis(dlatents, noise_mode=noise_mode)
    synth_image = (synth_image + 1) * 255 / 2  # [-1.0, 1.0] -> [0.0, 255.0]
    if to_np:
        synth_image = (
            synth_image.permute(0, 2, 3, 1).clamp(0, 255).to(torch.uint8).cpu().numpy()
        )  # NCWH => NWHC
    return synth_image


def get_w_from_seed(
    G,
    batch_sz: int,
    device: torch.device,
    truncation_psi: float,
    seed: Optional[int],
    centroids_path: Optional[str],
    class_idx: Optional[int],
) -> torch.Tensor:
    """Get the dlatent from a list of random seeds, using the truncation trick (this could be optional)"""

    if G.c_dim != 0:
        # sample random labels if no class idx is given
        if class_idx is None:
            class_indices = np.random.RandomState(seed).randint(
                low=0, high=G.c_dim, size=(batch_sz)
            )
            class_indices = torch.from_numpy(class_indices).to(device)
            w_avg = G.mapping.w_avg.index_select(0, class_indices)
        else:
            w_avg = G.mapping.w_avg[class_idx].unsqueeze(0).repeat(batch_sz, 1)
            class_indices = torch.full((batch_sz,), class_idx).to(device)

        labels = F.one_hot(class_indices, G.c_dim)

    else:
        w_avg = G.mapping.w_avg.unsqueeze(0)
        labels = None
        if class_idx is not None:
            print("Warning: --class is ignored when running an unconditional network")

    z = np.random.RandomState(seed).randn(batch_sz, G.z_dim)
    z = torch.from_numpy(z).to(device)
    w = G.mapping(z, labels)

    # multimodal truncation
    if centroids_path is not None:

        with dnnlib.util.open_url(centroids_path, verbose=False) as f:
            w_centroids = np.load(f)
        w_centroids = torch.from_numpy(w_centroids).to(device)
        w_centroids = w_centroids[None].repeat(batch_sz, 1, 1)

        # measure distances
        dist = torch.norm(w_centroids - w[:, :1], dim=2, p=2)
        w_avg = w_centroids[0].index_select(0, dist.argmin(1))

    w_avg = w_avg.unsqueeze(1).repeat(1, G.mapping.num_ws, 1)
    w = w_avg + (w - w_avg) * truncation_psi

    return w


def get_w_from_file(
    file: Union[str, os.PathLike], return_ext: bool = False
) -> Tuple[np.ndarray, Optional[str]]:
    """Get dlatent (w) from a .npy or .npz file"""
    filename, file_extension = os.path.splitext(file)
    assert file_extension in [
        ".npy",
        ".npz",
    ], f'"{file}" has wrong file format! Use either ".npy" or ".npz"'
    if file_extension == ".npy":
        r = (np.load(file), ".npy") if return_ext else np.load(file)
        return r
    r = (np.load(file)["w"], ".npz") if return_ext else np.load(file)["w"]
    return r


# ----------------------------------------------------------------------------


def save_config(
    ctx: click.Context, run_dir: Union[str, os.PathLike], save_name: str = "config.json"
) -> None:
    """Save the configuration stored in ctx.obj into a JSON file at the output directory."""
    with open(os.path.join(run_dir, save_name), "w") as f:
        json.dump(ctx.obj, f, indent=4, sort_keys=True)


# ----------------------------------------------------------------------------


def make_run_dir(
    outdir: Union[str, os.PathLike], desc: str, dry_run: bool = False
) -> str:
    """Reject modernity, return to automatically create the run dir."""
    # Pick output directory.
    prev_run_dirs = []
    if os.path.isdir(outdir):  # sanity check, but click.Path() should clear this one
        prev_run_dirs = [
            x for x in os.listdir(outdir) if os.path.isdir(os.path.join(outdir, x))
        ]
    prev_run_ids = [re.match(r"^\d+", x) for x in prev_run_dirs]
    prev_run_ids = [int(x.group()) for x in prev_run_ids if x is not None]
    cur_run_id = max(prev_run_ids, default=-1) + 1  # start with 00000
    run_dir = os.path.join(outdir, f"{cur_run_id:05d}-{desc}")
    assert not os.path.exists(run_dir)  # make sure it doesn't already exist

    # Don't create the dir if it's a dry-run
    if not dry_run:
        print("Creating output directory...")
        os.makedirs(run_dir)
    return run_dir



import os
import torch


def save_weights(model, filename, path="./saved_models"):
    if not os.path.isdir(path):
        os.makedirs(path)

    fpath = os.path.join(path, filename)
    torch.save(model.state_dict(), fpath)
    return


def save_checkpoint(model, optimizer, epoch, filename, root="./checkpoints"):
    if not os.path.isdir(root):
        os.makedirs(root)

    fpath = os.path.join(root, filename)
    torch.save(
        {
            "model": model.state_dict(),
            "optimizer": optimizer.state_dict(),
            "epoch": epoch,
        },
        fpath,
    )


def load_weights(model, filename, path="./saved_models"):
    fpath = os.path.join(path, filename)
    state_dict = torch.load(fpath)
    model.load_state_dict(state_dict)
    return model


def load_checkpoint(fpath, model, optimizer=None):
    ckpt = torch.load(fpath, map_location="cpu")
    if optimizer is None:
        optimizer = ckpt.get("optimizer", None)
    else:
        optimizer.load_state_dict(ckpt["optimizer"])
    epoch = ckpt["epoch"]

    if "model" in ckpt:
        ckpt = ckpt["model"]
    load_dict = {}
    for k, v in ckpt.items():
        if k.startswith("module."):
            k_ = k.replace("module.", "")
            load_dict[k_] = v
        else:
            load_dict[k] = v

    modified = {}  # backward compatibility to older naming of architecture blocks
    for k, v in load_dict.items():
        if k.startswith("adaptive_bins_layer.embedding_conv."):
            k_ = k.replace(
                "adaptive_bins_layer.embedding_conv.", "adaptive_bins_layer.conv3x3."
            )
            modified[k_] = v
            # del load_dict[k]

        elif k.startswith("adaptive_bins_layer.patch_transformer.embedding_encoder"):

            k_ = k.replace(
                "adaptive_bins_layer.patch_transformer.embedding_encoder",
                "adaptive_bins_layer.patch_transformer.embedding_convPxP",
            )
            modified[k_] = v
            # del load_dict[k]
        else:
            modified[k] = v  # else keep the original

    model.load_state_dict(modified)
    return (
        model,
        optimizer,
    )



from typing import List
import torch

from perceptor.models.owlvit.owlvit import OWLViTEncodings
from perceptor import models
from perceptor.losses.interface import LossInterface


class OWLViT(LossInterface):
    def __init__(self):
        """
        OWL-ViT zero-shot text-conditioned bounding box model.

        Loss is designed in such a way that only one example of the text prompt
        is expected to be found in the image.
        """
        super().__init__()
        self.model = models.OWLViT()
        self.encodings = None
        self.weights = None

    @property
    def device(self):
        return next(iter(self.model.parameters())).device

    def to(self, device):
        if self.encodings is not None:
            self.encodings.to(device)
        if self.weights is not None:
            self.weights.to(device)
        return super().to(device)

    def cuda(self):
        return self.to(torch.device("cuda"))

    def cpu(self):
        return self.to(torch.device("cpu"))

    def add_texts_(self, texts: List[str], weights=None):
        return self.add_encodings_(self.model.encode_texts([texts]), weights)

    def add_images_(self, images, weights=None):
        # return self.add_encodings_(self.model.encode_images(images), weights)
        raise NotImplementedError()

    def add_encodings_(
        self,
        encodings: OWLViTEncodings,
        weights=None,
    ):
        if isinstance(weights, list) or isinstance(weights, tuple):
            weights = torch.tensor(weights)
        elif weights is None:
            weights = torch.ones(len(encodings.texts))

        if self.encodings is None:
            self.encodings = encodings
            self.weights = torch.nn.Parameter(
                weights.to(self.device),
                requires_grad=False,
            )
        else:
            raise ValueError("OWLViT can only have one set of encodings")
        return self

    def forward(self, images, top_k=5):
        predictions = self.model(images, self.encodings)
        loss = torch.tensor(0.0, device=self.device)
        for label_index, weight in enumerate(self.weights):
            loss -= (
                predictions.logits[:, :, label_index]
                .flatten(start_dim=1)
                .log_softmax(dim=1)
                .sort(dim=1)[0][:, -top_k:]
                .mean()
                * weight
            )

        return loss * 0.01


def test_owlvit_loss():
    text_prompts = [
        "hello",
        "world",
    ]
    loss = OWLViT().add_texts_(text_prompts).cuda()
    loss(torch.zeros(3, 3, 480, 480).cuda())



from typing import TypeVar
from functools import wraps
import weakref


T = TypeVar("T")


def cache(model: T) -> T:
    cached = weakref.WeakValueDictionary()

    @wraps(model)
    def wrapper(*args, **kwargs):
        key = str(args) + str(kwargs)

        if key in cached:
            return cached[key]
        else:
            model_instance = model(*args, **kwargs)
            cached[key] = model_instance
            return model_instance

    return wrapper



import lpips

from .interface import LossInterface


class LPIPS(LossInterface):
    def __init__(self, name="squeeze", linear_layers=True, spatial=False):
        """
        LPIPS loss. Expects images of shape (batch_size, 3, height, width) between 0 and 1.

        Args:
            name (str): name of the loss. Available options: ["alex", "vgg", "squeeze"]
        """
        super().__init__()
        self.model = lpips.LPIPS(
            net=name, lpips=linear_layers, spatial=spatial, verbose=False
        )
        self.model.eval()
        self.model.requires_grad_(False)

    def forward(self, images_a, images_b):
        """
        Args:
            images_a: images of shape (batch_size, 3, height, width) between 0 and 1
            images_b: images of shape (batch_size, 3, height, width) between 0 and 1
        """
        return self.model(images_a, images_b, normalize=True)


def test_lpips():
    import torch

    model = LPIPS()
    images_a = torch.randn((1, 3, 256, 256))
    images_b = torch.randn((1, 3, 256, 256))
    model(images_a, images_b)



from typing import Optional
from contextlib import contextmanager
import copy
import torch
import lantern
from transformers import CLIPTokenizer, CLIPTextModel, logging
from diffusers import StableDiffusionPipeline, DDPMScheduler

from perceptor.utils import cache
from . import diffusion_space
from .predictions import Predictions
from .conditioning import Conditioning


class Model(torch.nn.Module):
    def __init__(
        self, name="CompVis/stable-diffusion-v1-4", fp16=False, auth_token=True
    ):
        """
        Args:
            name: The name of the model. Available models are:
                - CompVis/stable-diffusion-v1-4 (512x512)
                - CompVis/stable-diffusion-v1-3 (512x512)
                - CompVis/stable-diffusion-v1-2 (512x512)
                - CompVis/stable-diffusion-v1-1 (256x256)
        """
        super().__init__()
        self.name = name

        scheduler = DDPMScheduler(
            beta_start=0.00085, beta_end=0.012, beta_schedule="scaled_linear"
        )

        pipeline = StableDiffusionPipeline.from_pretrained(
            name,
            scheduler=scheduler,
            use_auth_token=auth_token,
            **dict(
                revision="fp16",
                torch_dtype=torch.float16,
            )
            if fp16
            else dict(),
        )

        self.vae = pipeline.vae
        self.unet = pipeline.unet
        self.feature_extractor = pipeline.feature_extractor
        self.scheduler = pipeline.scheduler

        self.schedule_alphas = torch.nn.Parameter(
            self.scheduler.alphas_cumprod.sqrt(), requires_grad=False
        )
        self.schedule_sigmas = torch.nn.Parameter(
            (1 - self.scheduler.alphas_cumprod).sqrt(),
            requires_grad=False,
        )

        self.vae_original_requires_grads = [
            parameter.requires_grad for parameter in self.vae.parameters()
        ]
        self.vae.eval()
        self.vae.requires_grad_(False)
        self.unet.eval()
        self.unet.requires_grad_(False)

    @property
    def device(self):
        return next(iter(self.parameters())).device

    @property
    def shape(self):
        return self.model.shape

    def schedule_indices(self, n_steps=500, from_index=999, to_index=0, rho=3.0):
        if from_index < to_index:
            raise ValueError("from_index must be greater than to_index")

        from_alpha, from_sigma = self.alphas(from_index), self.sigmas(from_index)
        to_alpha, to_sigma = self.alphas(to_index), self.sigmas(to_index)

        from_log_snr = torch.log(from_alpha**2 / from_sigma**2)
        to_log_snr = torch.log(to_alpha**2 / to_sigma**2)

        elucidated_from_sigma = (1 / from_log_snr.exp()).sqrt().clamp(max=150)
        elucidated_to_sigma = (1 / to_log_snr.exp()).sqrt().clamp(min=1e-3)

        ramp = torch.linspace(0, 1, n_steps + 1).to(self.device)
        min_inv_rho = elucidated_to_sigma ** (1 / rho)
        max_inv_rho = elucidated_from_sigma ** (1 / rho)
        sigmas = (max_inv_rho + ramp * (min_inv_rho - max_inv_rho)) ** rho
        target_log_snr = torch.log(torch.ones_like(sigmas) ** 2 / sigmas**2)

        schedule_log_snr = torch.log(
            self.schedule_alphas**2 / self.schedule_sigmas**2
        )

        assert target_log_snr.squeeze().ndim == 1
        assert schedule_log_snr.squeeze().ndim == 1
        schedule_indices = (
            (target_log_snr.squeeze()[:, None] - schedule_log_snr.squeeze()[None, :])
            .abs()
            .argmin(dim=1)
            .unique()
            .sort(descending=True)[0]
        )

        assert len(schedule_indices) >= n_steps * 0.9

        assert (schedule_indices[:-1] != schedule_indices[1:]).all()
        return torch.stack([schedule_indices[:-1], schedule_indices[1:]], dim=1)

    def encode(
        self, images: lantern.Tensor.dims("NCHW").float()
    ) -> lantern.Tensor.dims("NCHW"):
        _, _, h, w = images.shape
        if h % 32 != 0:
            raise Exception(f"Height must be divisible by 32, got {h}")
        if w % 32 != 0:
            raise Exception(f"Width must be divisible by 32, got {w}")
        return (
            0.18215
            * self.vae.encode(diffusion_space.encode(images.to(self.device))).sample()
        )

    def decode(
        self, latents: lantern.Tensor.dims("NCHW").float()
    ) -> lantern.Tensor.dims("NCHW"):
        return diffusion_space.decode(self.vae.decode(latents / 0.18215))

    @contextmanager
    def finetuneable_vae(self):
        """
        with diffusion_model.finetuneable_vae():
            images = diffusion_model.decode(latents)
        """
        state_dict = copy.deepcopy(self.vae.state_dict())
        try:
            for parameter, requires_grad in zip(
                self.vae.parameters(), self.vae_original_requires_grads
            ):
                parameter.requires_grad_(requires_grad)
            yield self
        finally:
            self.vae.load_state_dict(state_dict)
            self.vae.requires_grad_(False)

    @torch.cuda.amp.autocast()
    def latents(
        self, images: lantern.Tensor.dims("NCHW").float()
    ) -> lantern.Tensor.dims("NCHW"):
        return self.encode(images).float()

    @torch.cuda.amp.autocast()
    def images(
        self, latents: lantern.Tensor.dims("NCHW").float()
    ) -> lantern.Tensor.dims("NCHW"):
        return self.decode(latents).float()

    def random_diffused_latents(self, shape):
        n, c, h, w = shape
        if h % 8 != 0:
            raise ValueError("Height must be divisible by 32")
        if w % 8 != 0:
            raise ValueError("Width must be divisible by 32")
        return torch.randn((n, self.unet.in_channels, h // 8, w // 8)).to(self.device)

    def indices(self, indices):
        if isinstance(indices, float) or isinstance(indices, int):
            indices = torch.as_tensor(indices)
        if indices.ndim == 0:
            indices = indices[None]
        if indices.ndim != 1:
            raise ValueError("indices must be a scalar or a 1-dimensional tensor")
        return indices.long().to(self.device)

    def alphas(self, indices):
        return self.schedule_alphas[self.indices(indices)][:, None, None, None].to(
            self.device
        )

    def sigmas(self, indices):
        return self.schedule_sigmas[self.indices(indices)][:, None, None, None].to(
            self.device
        )

    @torch.cuda.amp.autocast()
    def predicted_noise(
        self,
        diffused_latents,
        from_indices,
        conditioning: Optional[Conditioning] = None,
    ):
        if conditioning is None:
            conditioning = self.conditioning()

        predicted_noise = self.unet(
            diffused_latents, self.indices(from_indices), conditioning.encodings
        )["sample"]
        return predicted_noise.float()

    def forward(self, diffused_latents, indices, conditioning=None):
        indices = self.indices(indices)
        return Predictions(
            from_diffused_latents=diffused_latents,
            from_indices=indices,
            predicted_noise=self.predicted_noise(
                diffused_latents, indices, conditioning
            ),
            schedule_alphas=self.schedule_alphas,
            schedule_sigmas=self.schedule_sigmas,
        )

    def predictions(self, diffused_latents, ts, conditioning=None):
        return self.forward(diffused_latents, ts, conditioning)

    def conditioning(self, texts=None, images=None, encodings=None):
        """
        Create a conditioning object from a list of texts. Unconditional is an empty string.
        """
        if texts is None and images is None and encodings is None:
            texts = [""]

        verbosity = logging.get_verbosity()
        logging.set_verbosity_error()
        tokenizer = CLIPTokenizer.from_pretrained("openai/clip-vit-large-patch14")
        text_encoder = CLIPTextModel.from_pretrained(
            "openai/clip-vit-large-patch14"
        ).to(self.device)
        logging.set_verbosity(verbosity)

        tokenized_text = tokenizer(
            texts,
            padding="max_length",
            max_length=tokenizer.model_max_length,
            truncation=True,
            return_tensors="pt",
        )
        text_encodings = text_encoder(tokenized_text.input_ids.to(self.device))[0]
        return Conditioning(encodings=text_encodings)

    def diffuse_latents(self, denoised_latents, indices, noise=None):
        indices = self.indices(indices)
        if noise is None:
            noise = torch.randn_like(denoised_latents)
        alphas, sigmas = self.alphas(indices), self.sigmas(indices)
        return denoised_latents * alphas + noise * sigmas


StableDiffusion: Model = cache(Model)


def test_stable_diffusion_step():
    from diffusers import DDIMScheduler

    torch.set_grad_enabled(False)
    device = torch.device("cuda")

    model_name = "CompVis/stable-diffusion-v1-4"
    batch_size = 1
    height = 512
    width = 512

    from_index = 999
    to_index = 998

    texts = ["painting of a dog"]

    diffusion_model = Model(model_name).to(device)

    diffused_latents = diffusion_model.random_diffused_latents((1, 3, 512, 512))

    conditioning = diffusion_model.conditioning(texts)
    predictions = diffusion_model.predictions(
        diffused_latents, from_index, conditioning
    )
    compare_next_diffused_latents = predictions.step(to_index)

    del diffusion_model

    # compare with diffusers
    scheduler = DDIMScheduler(
        beta_start=0.00085, beta_end=0.012, beta_schedule="scaled_linear"
    )
    scheduler.set_timesteps(1000)
    pipeline = StableDiffusionPipeline.from_pretrained(
        model_name,
        scheduler=scheduler,
        use_auth_token=True,
    ).to(device)

    latents_shape = (batch_size, pipeline.unet.in_channels, height // 8, width // 8)
    assert latents_shape == diffused_latents.shape

    tokenized_text = pipeline.tokenizer(
        texts,
        padding="max_length",
        max_length=pipeline.tokenizer.model_max_length,
        truncation=True,
        return_tensors="pt",
    )
    text_embeddings = pipeline.text_encoder(
        tokenized_text.input_ids.to(pipeline.device)
    )[0]

    assert text_embeddings.shape == conditioning.encodings.shape
    assert torch.allclose(
        text_embeddings[0, 0], conditioning.encodings[0, 0], atol=1e-3
    )
    assert torch.allclose(text_embeddings, conditioning.encodings, atol=1e-3)

    index = next(iter(pipeline.scheduler.timesteps))
    assert from_index == index

    predicted_noise = pipeline.unet(
        diffused_latents, index, encoder_hidden_states=text_embeddings
    )["sample"]

    assert torch.allclose(predictions.predicted_noise, predicted_noise, atol=5e-3)

    next_diffused_latents = pipeline.scheduler.step(
        predicted_noise, index, diffused_latents
    )["prev_sample"]

    assert torch.allclose(
        next_diffused_latents[0, 0, 0, 0],
        compare_next_diffused_latents[0, 0, 0, 0],
        atol=1e-3,
    )
    assert torch.allclose(
        next_diffused_latents, compare_next_diffused_latents, atol=1e-3
    )


def test_stable_diffusion():
    from tqdm import tqdm
    from perceptor import utils

    torch.set_grad_enabled(False)
    device = torch.device("cuda")

    diffusion_model = StableDiffusion().to(device)
    conditioning = diffusion_model.conditioning(["photograph of a playful cat"])
    unconditioned_conditioning = diffusion_model.conditioning()
    diffused_latents = diffusion_model.random_diffused_latents((1, 3, 512, 512)).to(
        device
    )

    progress_bar = tqdm(diffusion_model.schedule_indices(to_index=0, n_steps=50))
    for from_indices, to_indices in progress_bar:

        for _ in range(10):
            positive_predictions = diffusion_model.predictions(
                diffused_latents,
                from_indices,
                conditioning,
            )
            diffused_latents = positive_predictions.step(to_indices)

            unconditioned_predictions = diffusion_model.predictions(
                diffused_latents,
                to_indices,
                unconditioned_conditioning,
            )
            if from_indices.item() <= 700:
                diffused_latents = unconditioned_predictions.noisy_reverse_step(
                    from_indices
                )
            else:
                diffused_latents = unconditioned_predictions.reverse_step(from_indices)

        step_predictions = diffusion_model.predictions(
            diffused_latents,
            from_indices,
            conditioning,
        )
        diffused_latents = step_predictions.step(to_indices)

        utils.pil_image(
            diffusion_model.images(step_predictions.denoised_latents).clamp(0, 1)
        ).save("tests/stable_diffusion.png")

        progress_bar.set_postfix(
            dict(
                from_indices=from_indices.item(),
                to_indices=to_indices.item(),
            )
        )

    predictions = diffusion_model.predictions(
        diffused_latents,
        to_indices,
        conditioning,
    )

    utils.pil_image(
        diffusion_model.images(predictions.denoised_latents).clamp(0, 1)
    ).save("tests/stable_diffusion.png")


def test_stable_diffusion_init_image():
    import requests
    from PIL import Image
    import torch
    import torchvision.transforms.functional as TF
    from tqdm import tqdm
    from perceptor import utils

    image_url = "http://images.cocodataset.org/val2017/000000039769.jpg"
    init_image = TF.to_tensor(
        Image.open(requests.get(image_url, stream=True).raw).resize((512, 512))
    )[None].cuda()

    torch.set_grad_enabled(False)
    device = torch.device("cuda")

    from_index = 400

    diffusion_model = StableDiffusion().to(device)
    conditioning = diffusion_model.conditioning(["photograph of playful lions"])
    unconditioned_conditioning = diffusion_model.conditioning()
    diffused_latents = diffusion_model.diffuse_latents(
        diffusion_model.latents(init_image), from_index
    )

    for from_indices, to_indices in tqdm(
        diffusion_model.schedule_indices(from_index=from_index, to_index=0, n_steps=50)
    ):
        for _ in range(4):
            # upscaled guided forward, random backward
            unconditioned_predictions = diffusion_model.predictions(
                diffused_latents,
                from_indices,
                unconditioned_conditioning,
            )
            positive_predictions = diffusion_model.predictions(
                diffused_latents,
                from_indices,
                conditioning,
            )
            guided_predictions = unconditioned_predictions.classifier_free_guidance(
                positive_predictions, guidance_scale=7.0
            )
            diffused_latents = guided_predictions.resample(to_indices)

        predictions = diffusion_model.predictions(
            diffused_latents,
            from_indices,
            conditioning,
        )
        diffused_latents = predictions.step(to_indices)

    predictions = diffusion_model.predictions(
        diffused_latents,
        to_indices,
        conditioning,
    )

    utils.pil_image(
        diffusion_model.images(predictions.denoised_latents).clamp(0, 1)
    ).save("tests/stable_diffusion_init_image.png")



# -*- coding: utf-8 -*-
import pywt
import torch
import torch.nn as nn
from taming.modules.diffusionmodules.model import Decoder

from .pytorch_wavelets_utils import SFB2D, _SFB2D, prep_filt_sfb2d, mode_to_int


class DecoderDWT(nn.Module):
    def __init__(self, ddconfig, embed_dim):
        super().__init__()
        if ddconfig.out_ch != 12:
            ddconfig.out_ch = 12
        self.post_quant_conv = torch.nn.Conv2d(embed_dim, ddconfig['z_channels'], 1)
        self.decoder = Decoder(**ddconfig)
        self.idwt = DWTInverse(mode='zero', wave='db1')

    def forward(self, x):
        # x = self.post_quant_conv(x)
        freq = self.decoder(x)
        img = self.dwt_to_img(freq)
        return img

    def dwt_to_img(self, img):
        b, c, h, w = img.size()
        low = img[:, :3, :, :]
        high = img[:, 3:, :, :].view(b, 3, 3, h, w)
        return self.idwt((low, [high]))


class DWTInverse(nn.Module):
    """ Performs a 2d DWT Inverse reconstruction of an image

    Args:
        wave (str or pywt.Wavelet): Which wavelet to use
        C: deprecated, will be removed in future
    """

    def __init__(self, wave='db1', mode='zero', trace_model=False):
        super().__init__()
        if isinstance(wave, str):
            wave = pywt.Wavelet(wave)
        if isinstance(wave, pywt.Wavelet):
            g0_col, g1_col = wave.rec_lo, wave.rec_hi
            g0_row, g1_row = g0_col, g1_col
        else:
            if len(wave) == 2:
                g0_col, g1_col = wave[0], wave[1]
                g0_row, g1_row = g0_col, g1_col
            elif len(wave) == 4:
                g0_col, g1_col = wave[0], wave[1]
                g0_row, g1_row = wave[2], wave[3]
        # Prepare the filters
        filts = prep_filt_sfb2d(g0_col, g1_col, g0_row, g1_row)
        self.register_buffer('g0_col', filts[0])
        self.register_buffer('g1_col', filts[1])
        self.register_buffer('g0_row', filts[2])
        self.register_buffer('g1_row', filts[3])
        self.mode = mode
        self.trace_model = trace_model

    def forward(self, coeffs):
        """
        Args:
            coeffs (yl, yh): tuple of lowpass and bandpass coefficients, where:
              yl is a lowpass tensor of shape :math:`(N, C_{in}, H_{in}',
              W_{in}')` and yh is a list of bandpass tensors of shape
              :math:`list(N, C_{in}, 3, H_{in}'', W_{in}'')`. I.e. should match
              the format returned by DWTForward

        Returns:
            Reconstructed input of shape :math:`(N, C_{in}, H_{in}, W_{in})`

        Note:
            :math:`H_{in}', W_{in}', H_{in}'', W_{in}''` denote the correctly
            downsampled shapes of the DWT pyramid.

        Note:
            Can have None for any of the highpass scales and will treat the
            values as zeros (not in an efficient way though).
        """
        yl, yh = coeffs
        ll = yl
        mode = mode_to_int(self.mode)

        # Do a multilevel inverse transform
        for h in yh[::-1]:
            if h is None:
                h = torch.zeros(ll.shape[0], ll.shape[1], 3, ll.shape[-2],
                                ll.shape[-1], device=ll.device)

            # 'Unpad' added dimensions
            if ll.shape[-2] > h.shape[-2]:
                ll = ll[..., :-1, :]
            if ll.shape[-1] > h.shape[-1]:
                ll = ll[..., :-1]
            if not self.trace_model:
                ll = SFB2D.apply(ll, h, self.g0_col, self.g1_col, self.g0_row, self.g1_row, mode)
            else:
                ll = _SFB2D(ll, h, self.g0_col, self.g1_col, self.g0_row, self.g1_row, mode)
        return ll



import torch.nn as nn
from .common import *


class ResidualSequential(nn.Sequential):
    def __init__(self, *args):
        super(ResidualSequential, self).__init__(*args)

    def forward(self, x):
        out = super(ResidualSequential, self).forward(x)
        # print(x.size(), out.size())
        x_ = None
        if out.size(2) != x.size(2) or out.size(3) != x.size(3):
            diff2 = x.size(2) - out.size(2)
            diff3 = x.size(3) - out.size(3)
            # print(1)
            x_ = x[
                :,
                :,
                diff2 / 2 : out.size(2) + diff2 / 2,
                diff3 / 2 : out.size(3) + diff3 / 2,
            ]
        else:
            x_ = x
        return out + x_

    def eval(self):
        print(2)
        for m in self.modules():
            m.eval()
        exit()


def get_block(num_channels, norm_layer, act_fun):
    layers = [
        nn.Conv2d(num_channels, num_channels, 3, 1, 1, bias=False),
        norm_layer(num_channels, affine=True),
        act(act_fun),
        nn.Conv2d(num_channels, num_channels, 3, 1, 1, bias=False),
        norm_layer(num_channels, affine=True),
    ]
    return layers


class ResNet(nn.Module):
    def __init__(
        self,
        num_input_channels,
        num_output_channels,
        num_blocks,
        num_channels,
        need_residual=True,
        act_fun="LeakyReLU",
        need_sigmoid=True,
        norm_layer=nn.BatchNorm2d,
        pad="reflection",
    ):
        """
        pad = 'start|zero|replication'
        """
        super(ResNet, self).__init__()

        if need_residual:
            s = ResidualSequential
        else:
            s = nn.Sequential

        stride = 1
        # First layers
        layers = [
            # nn.ReplicationPad2d(num_blocks * 2 * stride + 3),
            conv(num_input_channels, num_channels, 3, stride=1, bias=True, pad=pad),
            act(act_fun),
        ]
        # Residual blocks
        # layers_residual = []
        for i in range(num_blocks):
            layers += [s(*get_block(num_channels, norm_layer, act_fun))]

        layers += [
            nn.Conv2d(num_channels, num_channels, 3, 1, 1),
            norm_layer(num_channels, affine=True),
        ]

        # if need_residual:
        #     layers += [ResidualSequential(*layers_residual)]
        # else:
        #     layers += [Sequential(*layers_residual)]

        # if factor >= 2:
        #     # Do upsampling if needed
        #     layers += [
        #         nn.Conv2d(num_channels, num_channels *
        #                   factor ** 2, 3, 1),
        #         nn.PixelShuffle(factor),
        #         act(act_fun)
        #     ]
        layers += [
            conv(num_channels, num_output_channels, 3, 1, bias=True, pad=pad),
            nn.Sigmoid(),
        ]
        self.model = nn.Sequential(*layers)

    def forward(self, input):
        return self.model(input)

    def eval(self):
        self.model.eval()



from .aesthetic_visual_assessment import AestheticVisualAssessment
from .blip import BLIP
from .clip import CLIP
from .cloob import CLOOB
from .lit import LiT
from .velocity_diffusion import VelocityDiffusion
from .lpips import LPIPS
from .super_resolution import SuperResolution, SuperResolutionDiscriminator
from .memorability import Memorability
from .open_clip import OpenCLIP
from .owlvit import OWLViT
from .resize import Resize
from .ruclip import RuCLIP
from .simulacra_aesthetic import SimulacraAesthetic
from .slip import SLIP
from .smoothness import Smoothness
from .spherical_distance import SphericalDistance
from .style_transfer import StyleTransfer



from collections import OrderedDict
import torch
from basicsr.utils.download_util import load_file_from_url
from torchvision import transforms

from perceptor.transforms import resize
from . import models
from .tokenizer import SimpleTokenizer

slip_models_ckpt = {
    "SLIP_VITS16": "slip_small_100ep.pt",
    "SLIP_VITB16": "slip_base_100ep.pt",
    "SLIP_VITL16": "slip_large_100ep.pt",
    "CLIP_VITS16": "clip_small_25ep.pt",
    "CLIP_VITB16": "clip_base_25ep.pt",
    "CLIP_VITL16": "clip_large_25ep.pt",
    "SLIP_CC3M": "slip_base_cc3m_40ep.pt",
    "SLIP_CC12M": "slip_base_cc12m_35ep.pt",
}


def normalize(img, input_range=None):
    if input_range is None:
        minv = img.min()
    else:
        minv = input_range[0]
    img = img - minv

    if input_range is None:
        maxv = img.max()
    else:
        maxv = input_range[1] - minv

    if maxv != 0:
        img = img / maxv

    return img


def adjust_range(img, out_range, input_range=None):
    img = normalize(img, input_range=input_range)
    img = img * (out_range[1] - out_range[0])
    img = img + out_range[0]
    return img


class SLIP_Base(torch.nn.Module):
    def __init__(self, model_name):
        super().__init__()
        self.input_resolution = 224
        ckpt_file = slip_models_ckpt[model_name]
        ckpt_path = f"models/{ckpt_file}"
        load_file_from_url(
            f"https://dl.fbaipublicfiles.com/slip/{ckpt_file}",
            "models",
        )

        self.preprocess_transform = transforms.Compose(
            [
                # transforms.Resize(224),
                # transforms.CenterCrop(224),
                transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
            ]
        )

        self.tokenizer = SimpleTokenizer()

        ckpt = torch.load(ckpt_path, map_location="cpu")
        state_dict = OrderedDict()
        for k, v in ckpt["state_dict"].items():
            state_dict[k.replace("module.", "")] = v

        # create model
        old_args = ckpt["args"]
        old_args.model = model_name
        # these two are the same model on different training data...
        if old_args.model == "SLIP_CC3M" or old_args.model == "SLIP_CC12M":
            old_args.model = "SLIP_VITB16"

        self.model = getattr(models, old_args.model)(
            rand_embed=False,
            ssl_mlp_dim=old_args.ssl_mlp_dim,
            ssl_emb_dim=old_args.ssl_emb_dim,
        )
        self.model.requires_grad_(False).eval()
        self.model.load_state_dict(state_dict, strict=True)
        n_params = sum(p.numel() for p in self.model.parameters())
        print("Loaded perceptor %s: %.2fM params" % (model_name, (n_params / 1000000)))

    def preprocess(self, imgs, input_range=None):
        imgs = adjust_range(imgs, [0.0, 1.0], input_range=input_range)
        return self.preprocess_transform(imgs)

    def encode_image(self, imgs, input_range=None, apply_preprocess=True):
        if apply_preprocess:
            imgs = resize(
                imgs,
                out_shape=(self.input_resolution, self.input_resolution),
            )
            imgs = self.preprocess(imgs, input_range=input_range)
        image_features = self.model.encode_image(imgs)
        image_features = image_features / image_features.norm(dim=-1, keepdim=True)
        return image_features

    def encode_text(self, texts):
        texts = self.tokenizer(texts)
        texts = texts.view(-1, 77).contiguous()
        text_embeddings = self.model.encode_text(texts)
        return text_embeddings

    def encode_texts(self, texts):
        texts = self.tokenizer(texts)
        texts = texts.view(-1, 77).contiguous()
        text_embeddings = self.model.encode_text(texts)
        text_embeddings = text_embeddings / text_embeddings.norm(dim=-1, keepdim=True)
        return text_embeddings.unsqueeze(1)


def get_slip_perceptor(model_name):
    return SLIP_Base(model_name)



from .skip import skip
from .texture_nets import get_texture_nets
from .resnet import ResNet
from .unet import UNet

import torch.nn as nn


def get_net(
    input_depth,
    NET_TYPE,
    pad,
    upsample_mode,
    n_channels=3,
    act_fun="LeakyReLU",
    skip_n33d=128,
    skip_n33u=128,
    skip_n11=4,
    num_scales=5,
    downsample_mode="stride",
):
    if NET_TYPE == "ResNet":
        # TODO
        net = ResNet(input_depth, 3, 10, 16, 1, nn.BatchNorm2d, False)
    elif NET_TYPE == "skip":
        net = skip(
            input_depth,
            n_channels,
            num_channels_down=[skip_n33d] * num_scales
            if isinstance(skip_n33d, int)
            else skip_n33d,
            num_channels_up=[skip_n33u] * num_scales
            if isinstance(skip_n33u, int)
            else skip_n33u,
            num_channels_skip=[skip_n11] * num_scales
            if isinstance(skip_n11, int)
            else skip_n11,
            upsample_mode=upsample_mode,
            downsample_mode=downsample_mode,
            need_sigmoid=True,
            need_bias=True,
            pad=pad,
            act_fun=act_fun,
        )

    elif NET_TYPE == "texture_nets":
        net = get_texture_nets(
            inp=input_depth, ratios=[32, 16, 8, 4, 2, 1], fill_noise=False, pad=pad
        )

    elif NET_TYPE == "UNet":
        net = UNet(
            num_input_channels=input_depth,
            num_output_channels=3,
            feature_scale=4,
            more_layers=0,
            concat_x=False,
            upsample_mode=upsample_mode,
            pad=pad,
            norm_layer=nn.BatchNorm2d,
            need_sigmoid=True,
            need_bias=True,
        )
    elif NET_TYPE == "identity":
        assert input_depth == 3
        net = nn.Sequential()
    else:
        assert False

    return net


def get_hq_skip_net(
    input_depth,
    pad="reflection",
    upsample_mode="cubic",
    n_channels=3,
    act_fun="LeakyReLU",
    skip_n33d=192,
    skip_n33u=192,
    skip_n11=4,
    num_scales=6,
    downsample_mode="cubic",
    decorr_rgb=True,
    offset_groups=4,
    offset_type="1x1",
    sigmoid=True,
):
    """Constructs and returns a skip network with higher quality default settings, including
    deformable convolutions (can be slow, disable with offset_groups=0). Further
    improvements can be seen by setting offset_type to 'full', but then you may have to
    reduce the learning rate of the offset layers to ~1/10 of the rest of the layers. See
    the get_offset_params() and get_non_offset_params() functions to construct the
    parameter groups."""
    net = skip(
        input_depth,
        n_channels,
        num_channels_down=[skip_n33d] * num_scales
        if isinstance(skip_n33d, int)
        else skip_n33d,
        num_channels_up=[skip_n33u] * num_scales
        if isinstance(skip_n33u, int)
        else skip_n33u,
        num_channels_skip=[skip_n11] * num_scales
        if isinstance(skip_n11, int)
        else skip_n11,
        upsample_mode=upsample_mode,
        downsample_mode=downsample_mode,
        need_sigmoid=sigmoid,
        need_bias=True,
        decorr_rgb=decorr_rgb,
        pad=pad,
        act_fun=act_fun,
        offset_groups=offset_groups,
        offset_type=offset_type,
    )
    return net


def get_offset_params(net):
    """Returns an iterable of parameters of layers that output offsets for deformable
    convolutions (for setting their learning rate lower than the rest).
    Example:
        >>> params = [{'params': get_non_offset_params(net), 'lr': lr},
        >>>           {'params': get_offset_params(net), 'lr': lr / 10}]
        >>> opt = optim.Adam(params)
    """
    return [p for n, p in net.named_parameters() if "offset_branch" in n]


def get_non_offset_params(net):
    """Returns an iterable of parameters of layers that do not output offsets for
    deformable convolutions.
    Example:
        >>> params = [{'params': get_non_offset_params(net), 'lr': lr},
        >>>           {'params': get_offset_params(net), 'lr': lr / 10}]
        >>> opt = optim.Adam(params)
    """
    return [p for n, p in net.named_parameters() if "offset_branch" not in n]



# https://github.com/mlomnitz/DiffJPEG
import itertools
import numpy as np
import torch
import torch.nn as nn
from .utils import c_table, y_table, diff_round


class y_dequantize(nn.Module):
    """Dequantize Y channel
    Inputs:
        image(tensor): batch x height x width
        factor(float): compression factor
    Outputs:
        image(tensor): batch x height x width
    """

    def __init__(self, factor=1):
        super(y_dequantize, self).__init__()
        self.y_table = nn.Parameter(y_table(), requires_grad=False)
        self.factor = factor

    def forward(self, image):
        return image * (self.y_table * self.factor)


class c_dequantize(nn.Module):
    """Dequantize CbCr channel
    Inputs:
        image(tensor): batch x height x width
        factor(float): compression factor
    Outputs:
        image(tensor): batch x height x width
    """

    def __init__(self, factor=1):
        super(c_dequantize, self).__init__()
        self.factor = factor
        self.c_table = nn.Parameter(c_table(), requires_grad=False)

    def forward(self, image):
        return image * (self.c_table * self.factor)


class idct_8x8(nn.Module):
    """Inverse discrete Cosine Transformation
    Input:
        dcp(tensor): batch x height x width
    Output:
        image(tensor): batch x height x width
    """

    def __init__(self):
        super(idct_8x8, self).__init__()
        alpha = np.array([1.0 / np.sqrt(2)] + [1] * 7)
        self.alpha = nn.Parameter(
            torch.from_numpy(np.outer(alpha, alpha)).float(), requires_grad=False
        )
        tensor = np.zeros((8, 8, 8, 8), dtype=np.float32)
        for x, y, u, v in itertools.product(range(8), repeat=4):
            tensor[x, y, u, v] = np.cos((2 * u + 1) * x * np.pi / 16) * np.cos(
                (2 * v + 1) * y * np.pi / 16
            )
        self.tensor = nn.Parameter(
            torch.from_numpy(tensor).float(), requires_grad=False
        )

    def forward(self, image):

        image = image * self.alpha
        result = 0.25 * torch.tensordot(image, self.tensor, dims=2) + 128
        result.view(image.shape)
        return result


class block_merging(nn.Module):
    """Merge pathces into image
    Inputs:
        patches(tensor) batch x height*width/64, height x width
        height(int)
        width(int)
    Output:
        image(tensor): batch x height x width
    """

    def __init__(self):
        super(block_merging, self).__init__()

    def forward(self, patches, height, width):
        k = 8
        batch_size = patches.shape[0]
        image_reshaped = patches.view(batch_size, height // k, width // k, k, k)
        image_transposed = image_reshaped.permute(0, 1, 3, 2, 4)
        return image_transposed.contiguous().view(batch_size, height, width)


class chroma_upsampling(nn.Module):
    """Upsample chroma layers
    Input:
        y(tensor): y channel image
        cb(tensor): cb channel
        cr(tensor): cr channel
    Ouput:
        image(tensor): batch x height x width x 3
    """

    def __init__(self):
        super(chroma_upsampling, self).__init__()

    def forward(self, y, cb, cr):
        def repeat(x, k=2):
            height, width = x.shape[1:3]
            x = x.unsqueeze(-1)
            x = x.repeat(1, 1, k, k)
            x = x.view(-1, height * k, width * k)
            return x

        cb = repeat(cb)
        cr = repeat(cr)

        return torch.cat([y.unsqueeze(3), cb.unsqueeze(3), cr.unsqueeze(3)], dim=3)


class ycbcr_to_rgb_jpeg(nn.Module):
    """Converts YCbCr image to RGB JPEG
    Input:
        image(tensor): batch x height x width x 3
    Outpput:
        result(tensor): batch x 3 x height x width
    """

    def __init__(self):
        super(ycbcr_to_rgb_jpeg, self).__init__()

        matrix = np.array(
            [[1.0, 0.0, 1.402], [1, -0.344136, -0.714136], [1, 1.772, 0]],
            dtype=np.float32,
        ).T
        self.shift = nn.Parameter(
            torch.tensor([0, -128.0, -128.0]), requires_grad=False
        )
        self.matrix = nn.Parameter(torch.from_numpy(matrix), requires_grad=False)

    def forward(self, image):
        result = torch.tensordot(image + self.shift, self.matrix, dims=1)
        result.view(image.shape)
        return result.permute(0, 3, 1, 2)


class decompress_jpeg(nn.Module):
    """Full JPEG decompression algortihm
    Input:
        compressed(dict(tensor)): batch x h*w/64 x 8 x 8
        rounding(function): rounding function to use
        factor(float): Compression factor
    Ouput:
        image(tensor): batch x 3 x height x width
    """

    def __init__(self, height, width, rounding=diff_round, factor=1):
        super(decompress_jpeg, self).__init__()
        self.c_dequantize = c_dequantize(factor=factor)
        self.y_dequantize = y_dequantize(factor=factor)
        self.idct = idct_8x8()
        self.merging = block_merging()
        self.chroma = chroma_upsampling()
        self.colors = ycbcr_to_rgb_jpeg()

        self.height, self.width = height, width

    def forward(self, y, cb, cr):
        components = {"y": y, "cb": cb, "cr": cr}
        for k in components.keys():
            if k in ("cb", "cr"):
                comp = self.c_dequantize(components[k])
                height, width = int(self.height / 2), int(self.width / 2)
            else:
                comp = self.y_dequantize(components[k])
                height, width = self.height, self.width
            comp = self.idct(comp)
            components[k] = self.merging(comp, height, width)
            #
        image = self.chroma(components["y"], components["cb"], components["cr"])
        image = self.colors(image)

        image = torch.min(
            255 * torch.ones_like(image), torch.max(torch.zeros_like(image), image)
        )
        return image / 255



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['perceptor',
 'perceptor.drawers',
 'perceptor.drawers.diffusion',
 'perceptor.drawers.jpeg',
 'perceptor.drawers.raw',
 'perceptor.drawers.raw.init',
 'perceptor.drawers.rudalle',
 'perceptor.losses',
 'perceptor.losses.clip',
 'perceptor.losses.super_resolution',
 'perceptor.models',
 'perceptor.models.adabins_depth',
 'perceptor.models.adabins_depth.models',
 'perceptor.models.blip',
 'perceptor.models.cloob',
 'perceptor.models.deep_image_prior',
 'perceptor.models.glide_clip',
 'perceptor.models.guided_diffusion',
 'perceptor.models.latent_diffusion',
 'perceptor.models.latent_diffusion.finetuned',
 'perceptor.models.latent_diffusion.finetuned.encoders',
 'perceptor.models.latent_diffusion.ldm',
 'perceptor.models.latent_diffusion.ldm.models',
 'perceptor.models.latent_diffusion.ldm.models.diffusion',
 'perceptor.models.latent_diffusion.ldm.modules',
 'perceptor.models.latent_diffusion.ldm.modules.diffusionmodules',
 'perceptor.models.latent_diffusion.ldm.modules.distributions',
 'perceptor.models.latent_diffusion.ldm.modules.encoders',
 'perceptor.models.midas_depth',
 'perceptor.models.monster_diffusion',
 'perceptor.models.monster_diffusion.base',
 'perceptor.models.owlvit',
 'perceptor.models.ruclip',
 'perceptor.models.simulacra_aesthetic',
 'perceptor.models.slip',
 'perceptor.models.stable_diffusion',
 'perceptor.models.stylegan_xl',
 'perceptor.models.stylegan_xl.torch_utils',
 'perceptor.models.stylegan_xl.torch_utils.ops',
 'perceptor.models.super_resolution',
 'perceptor.models.velocity_diffusion',
 'perceptor.models.vgg',
 'perceptor.transforms',
 'perceptor.transforms.resize',
 'perceptor.utils']

package_data = \
{'': ['*'],
 'perceptor.losses.clip': ['vectors/*'],
 'perceptor.models.blip': ['configs/*'],
 'perceptor.models.cloob': ['pretrained_configs/*'],
 'perceptor.models.latent_diffusion': ['configs/autoencoder/*',
                                       'configs/latent-diffusion/*',
                                       'models/first_stage_models/kl-f16/*',
                                       'models/first_stage_models/kl-f32/*',
                                       'models/first_stage_models/kl-f4/*',
                                       'models/first_stage_models/kl-f8/*',
                                       'models/first_stage_models/vq-f16/*',
                                       'models/first_stage_models/vq-f4-noattn/*',
                                       'models/first_stage_models/vq-f4/*',
                                       'models/first_stage_models/vq-f8-n256/*',
                                       'models/first_stage_models/vq-f8/*',
                                       'models/ldm/bsr_sr/*',
                                       'models/ldm/celeba256/*',
                                       'models/ldm/cin256/*',
                                       'models/ldm/ffhq256/*',
                                       'models/ldm/inpainting_big/*',
                                       'models/ldm/layout2img-openimages256/*',
                                       'models/ldm/lsun_beds256/*',
                                       'models/ldm/lsun_churches256/*',
                                       'models/ldm/semantic_synthesis256/*',
                                       'models/ldm/semantic_synthesis512/*',
                                       'models/ldm/text2img256/*'],
 'perceptor.models.simulacra_aesthetic': ['weights/*']}

install_requires = \
['Pillow>=9.0.1,<10.0.0',
 'basicsr>=1.4.2,<2.0.0',
 'diffusers>=0.2.4,<0.3.0',
 'dill>=0.3.4,<0.4.0',
 'einops>=0.4.0,<0.5.0',
 'ftfy>=6.0.3,<7.0.0',
 'imageio>=2.14.1,<3.0.0',
 'kornia>=0.6.3,<0.7.0',
 'lpips>=0.1.4,<0.2.0',
 'more-itertools>=8.12.0,<9.0.0',
 'ninja>=1.10.2,<2.0.0',
 'numpy>=1.22.2,<2.0.0',
 'omegaconf>=2.1.1,<3.0.0',
 'open-clip-torch>=2.0.2,<3.0.0',
 'opencv-python>=4.5.4,<5.0.0,!=4.6.0.66',
 'pytorch-lantern>=0.12.0,<0.13.0',
 'pytorch-zero-lit>=0.2.2,<0.3.0',
 'resmem>=1.1.4,<2.0.0',
 'taming-transformers-rom1504>=0.0.6,<0.0.7',
 'timm>=0.5.4,<0.6.0',
 'torch>=1.11.0,<2.0.0',
 'torchvision>=0.12.0,<0.13.0',
 'tqdm>=4.62.3,<5.0.0',
 'transformers>=4.21.1,<5.0.0',
 'youtokentome>=1.0.6,<2.0.0']

setup_kwargs = {
    'name': 'perceptor',
    'version': '0.5.15',
    'description': 'Modular image generation library',
    'long_description': '# perceptor\n\nModular image generation library.\n\n## Install\n\n```\npoetry add perceptor\n```\n\nOr, for the old timers:\n\n```\npip install perceptor\n```\n\n## API\n\nShortlist of available features.\n\n```python\nperceptor\n  .drawers\n    .BruteDiffusion\n    .DeepImagePrior\n    .JPEG\n    .Raw\n    .RuDALLE\n    .StyleGANXL\n  .models\n    .AdaBinsDepth\n    .DeepImagePrior\n    .VelocityDiffusion (yfcc_2, yfcc_1, cc12m_1_cfg, wikiart)\n    .latent_diffusion\n      .Text2Image\n      .Face\n      .SuperResolution\n    .GuidedDiffusion (openai, pixelart)\n    .MidasDepth\n    .MonsterDiffusion (all, tinyhero)\n    .StyleGANXL\n    .RuDALLE\n    .StableDiffusion\n    .SuperResolution\n  .losses\n    # Text-image similarity\n    .BLIP\n    .CLIP\n    .CLOOB\n    .LiT\n    .GlideCLIP\n    .OpenCLIP\n    .OWLViT\n    .RuCLIP\n    .SLIP\n    # Other\n    .AestheticVisualAssessment\n    .LPIPS\n    .Memorability\n    .MidasDepth\n    .SimulacraAesthetic\n    .Smoothness\n    .StyleTransfer\n  .transforms\n    .clamp_with_grad\n    .resize\n    .SuperResolution\n  .utils\n    .gradient_checkpoint\n```\n',
    'author': 'Richard Löwenström',
    'author_email': 'samedii@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/samedii/perceptor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from math import sqrt, log

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import einsum
from einops import rearrange
from taming.modules.diffusionmodules.model import Encoder, Decoder

from .decoder_dwt import DecoderDWT


class VQGanGumbelVAE(torch.nn.Module):

    def __init__(self, config, dwt=False):
        super().__init__()
        model = GumbelVQ(
            ddconfig=config.model.params.ddconfig,
            n_embed=config.model.params.n_embed,
            embed_dim=config.model.params.embed_dim,
            kl_weight=config.model.params.kl_weight,
            dwt=dwt,
        )
        self.model = model
        self.num_layers = int(log(config.model.params.ddconfig.attn_resolutions[0]) / log(2))
        self.image_size = 256
        self.num_tokens = config.model.params.n_embed

    @torch.no_grad()
    def get_codebook_indices(self, img):
        img = (2 * img) - 1
        _, _, [_, _, indices] = self.model.encode(img)
        return rearrange(indices, 'b h w -> b (h w)')

    def decode(self, img_seq):
        b, n = img_seq.shape
        one_hot_indices = torch.nn.functional.one_hot(img_seq, num_classes=self.num_tokens).float()
        z = (one_hot_indices @ self.model.quantize.embed.weight)
        z = rearrange(z, 'b (h w) c -> b c h w', h=int(sqrt(n)))
        img = self.model.decode(z)
        img = (img.clamp(-1., 1.) + 1) * 0.5
        return img


class GumbelQuantize(nn.Module):
    """
    credit to @karpathy: https://github.com/karpathy/deep-vector-quantization/blob/main/model.py (thanks!)
    Gumbel Softmax trick quantizer
    Categorical Reparameterization with Gumbel-Softmax, Jang et al. 2016
    https://arxiv.org/abs/1611.01144
    """

    def __init__(self, num_hiddens, embedding_dim, n_embed, straight_through=True,
                 kl_weight=5e-4, temp_init=1.0, use_vqinterface=True):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.n_embed = n_embed
        self.straight_through = straight_through
        self.temperature = temp_init
        self.kl_weight = kl_weight
        self.proj = nn.Conv2d(num_hiddens, n_embed, 1)
        self.embed = nn.Embedding(self.n_embed, self.embedding_dim)
        self.use_vqinterface = use_vqinterface

    def forward(self, z, temp=None, return_logits=False):
        hard = self.straight_through if self.training else True
        temp = self.temperature if temp is None else temp
        logits = self.proj(z)
        soft_one_hot = F.gumbel_softmax(logits, tau=temp, dim=1, hard=hard)
        z_q = einsum('b n h w, n d -> b d h w', soft_one_hot, self.embed.weight)
        # + kl divergence to the prior loss
        qy = F.softmax(logits, dim=1)
        diff = self.kl_weight * torch.sum(qy * torch.log(qy * self.n_embed + 1e-10), dim=1).mean()
        ind = soft_one_hot.argmax(dim=1)
        if self.use_vqinterface:
            if return_logits:
                return z_q, diff, (None, None, ind), logits
            return z_q, diff, (None, None, ind)
        return z_q, diff, ind


class GumbelVQ(nn.Module):

    def __init__(self, ddconfig, n_embed, embed_dim, dwt=False, kl_weight=1e-8):
        super().__init__()
        z_channels = ddconfig['z_channels']
        self.dwt = dwt
        self.encoder = Encoder(**ddconfig)
        self.decoder = DecoderDWT(ddconfig, embed_dim) if dwt else Decoder(**ddconfig)
        self.quantize = GumbelQuantize(z_channels, embed_dim, n_embed=n_embed, kl_weight=kl_weight, temp_init=1.0)
        self.quant_conv = torch.nn.Conv2d(ddconfig['z_channels'], embed_dim, 1)
        self.post_quant_conv = torch.nn.Conv2d(embed_dim, ddconfig['z_channels'], 1)

    def encode(self, x):
        h = self.encoder(x)
        h = self.quant_conv(h)
        quant, emb_loss, info = self.quantize(h)
        return quant, emb_loss, info

    def decode(self, quant):
        if self.dwt:
            quant = self.decoder.post_quant_conv(quant)
        else:
            quant = self.post_quant_conv(quant)
        dec = self.decoder(quant)
        return dec



import torch
import torch.nn.functional as F

from perceptor import utils
from .slip_base import get_slip_perceptor


@utils.cache
class SLIP(torch.nn.Module):
    def __init__(self, name):
        """
        Args:
            name: name of the slip model. Available models are:
                - SLIP_VITS16
                - SLIP_VITB16
                - SLIP_VITL16
                - CLIP_VITS16
                - CLIP_VITB16
                - CLIP_VITL16
                - SLIP_CC3M
                - SLIP_CC12M
        """
        super().__init__()
        self.name = name
        self.model = get_slip_perceptor(name).requires_grad_(False).eval()

    @property
    def device(self):
        return next(iter(self.parameters()))

    def encode_texts(self, text_prompts):
        return F.normalize(self.model.encode_text(text_prompts))

    def encode_images(self, images):
        return F.normalize(self.model.encode_image(images.to(self.device)))

    def forward(self, _):
        raise NotImplementedError



# https://github.com/ChaoningZhang/Pseudo-Differentiable-JPEG
# https://github.com/mlomnitz/DiffJPEG/blob/master/utils.py
import numpy as np
import torch


def y_table():
    return torch.from_numpy(
        np.array(
            [
                [16, 11, 10, 16, 24, 40, 51, 61],
                [12, 12, 14, 19, 26, 58, 60, 55],
                [14, 13, 16, 24, 40, 57, 69, 56],
                [14, 17, 22, 29, 51, 87, 80, 62],
                [18, 22, 37, 56, 68, 109, 103, 77],
                [24, 35, 55, 64, 81, 104, 113, 92],
                [49, 64, 78, 87, 103, 121, 120, 101],
                [72, 92, 95, 98, 112, 100, 103, 99],
            ],
            dtype=np.float32,
        ).T
    )


def c_table():
    c_table = np.empty((8, 8), dtype=np.float32)
    c_table.fill(99)
    c_table[:4, :4] = np.array(
        [[17, 18, 24, 47], [18, 21, 26, 66], [24, 26, 56, 99], [47, 66, 99, 99]]
    ).T
    return torch.from_numpy(c_table)


def diff_round(x):
    """Differentiable rounding function
    Input:
        x(tensor)
    Output:
        x(tensor)
    """
    return torch.round(x) + (x - torch.round(x)) ** 3


def quality_to_factor(quality):
    """Calculate factor corresponding to quality
    Input:
        quality(float): Quality for jpeg compression
    Output:
        factor(float): Compression factor
    """
    if quality < 50:
        quality = 5000.0 / quality
    else:
        quality = 200.0 - quality * 2
    return quality / 100.0



import torch
import torch.nn.functional as F
from torchvision import transforms
from basicsr.utils.download_util import load_file_from_url

from perceptor import models
from perceptor.losses.interface import LossInterface


class BLIP(LossInterface):
    def __init__(self, name="model_base_retrieval_flickr"):
        """
        Args:
            name (str): name of the blip model. Available models are:
                - model_base_retrieval_coco
                - model_large_retrieval_coco
                - model_base_retrieval_flickr
                - model_large_retrieval_flickr
                - model_large
                - model*_base
                - model_base
                - model_base_capfilt_large
        """
        super().__init__()
        self.name = name
        self.model = models.BLIP(name)

        self.encodings = None
        self.weights = None

    @property
    def device(self):
        return next(iter(self.model.parameters())).device

    def add_texts_(self, texts, weights=None):
        text_encodings = self.model.encode_texts(texts)
        return self.add_encodings_(text_encodings, weights)

    def add_images_(self, images, weights=None):
        image_encodings = self.model.encode_images(images)
        return self.add_encodings_(image_encodings, weights)

    def add_encodings_(self, encodings, weights=None):
        if isinstance(weights, list) or isinstance(weights, tuple):
            weights = torch.tensor(weights)
        elif weights is None:
            weights = torch.ones_like(encodings[:, 0])

        if self.encodings is None:
            self.encodings = torch.nn.Parameter(
                encodings.to(self.device), requires_grad=False
            )
            self.weights = torch.nn.Parameter(
                weights.to(self.device),
                requires_grad=False,
            )
        else:
            self.encodings = torch.nn.Parameter(
                torch.cat([self.encodings, encodings.to(self.device)]),
                requires_grad=False,
            )
            self.weights = torch.nn.Parameter(
                torch.cat([self.weights, weights.to(self.device)]),
                requires_grad=False,
            )
        return self

    def forward(self, images):
        return (
            self.model.image_text_contrastive_spherical_distance(
                self.model.encode_images(images), self.encodings
            )
            * self.weights[:, None]
        ).mean()


def test_blip_loss():
    loss = (
        BLIP().add_texts_(["hello", "world"]).add_images_(torch.randn(1, 3, 256, 256))
    )
    loss(torch.randn(1, 3, 256, 256))



import math
import pickle
import torch
from torch import nn
from torch.nn import functional as F
from torchvision import transforms
import open_clip


class PositionalEmbedding(nn.Module):
    def __init__(self, seq_len, d_model):
        super().__init__()
        self.weight = nn.Parameter(torch.randn([seq_len, d_model]))

    def forward(self, x):
        return x + self.weight


class GELU(nn.Module):
    def __init__(self, approximate=True):
        super().__init__()
        self.approximate = approximate

    def forward(self, x):
        if self.approximate:
            return (
                x * (1 + torch.tanh((2 / math.pi) ** 0.5 * (x + 0.047715 * x**3))) / 2
            )
        return F.gelu(x)


class SelfAttention(nn.Module):
    def __init__(self, d_model, n_heads=1):
        super().__init__()
        assert d_model % n_heads == 0
        self.norm = nn.LayerNorm(d_model)
        self.query = nn.Linear(d_model, d_model)
        self.key = nn.Linear(d_model, d_model)
        self.value = nn.Linear(d_model, d_model)
        self.out = nn.Linear(d_model, d_model)
        self.n_heads = n_heads

    def forward(self, x, padding_mask=None):
        n, s, d = x.shape
        head_size = d // self.n_heads
        x = self.norm(x)
        q = self.query(x).view([*x.shape[:-1], self.n_heads, head_size])
        k = self.key(x).view([*x.shape[:-1], self.n_heads, head_size])
        v = self.value(x).view([*x.shape[:-1], self.n_heads, head_size])
        attn_logits = torch.einsum("...thd,...Thd->...htT", q, k) / head_size**0.5
        if padding_mask is not None:
            mask = padding_mask[:, None, :, None]
            attn_logits = torch.where(mask, attn_logits, attn_logits.new_tensor(-1e30))
        attn_weights = attn_logits.softmax(-1)
        attn = torch.einsum("...htT,...Thd->...thd", attn_weights, v)
        attn_vec = attn.reshape(x.shape)
        return self.out(attn_vec)


class FeedForward(nn.Module):
    def __init__(self, d_model, d_ff):
        super().__init__()
        self.norm = nn.LayerNorm(d_model)
        self.linear_0 = nn.Linear(d_model, d_ff)
        self.act = nn.GELU()
        self.linear_1 = nn.Linear(d_ff, d_model)

    def forward(self, x):
        x = self.norm(x)
        x = self.linear_0(x)
        x = self.act(x)
        x = self.linear_1(x)
        return x


class TransformerEncoderLayer(nn.Module):
    def __init__(self, d_model, d_ff, n_heads):
        super().__init__()
        self.attn = SelfAttention(d_model, n_heads)
        self.ff = FeedForward(d_model, d_ff)

    def __call__(self, x, padding_mask=None):
        x = x + self.attn(x, padding_mask)
        x = x + self.ff(x)
        return x


class TextEncoder(nn.Module):
    def __init__(self, d_embed, n_layers, d_model, seq_len, n_heads, vocab_size):
        super().__init__()
        self.d_embed = d_embed
        self.n_layers = n_layers
        self.d_model = d_model
        self.d_ff = self.d_model * 4
        self.seq_len = seq_len
        self.n_heads = n_heads
        self.vocab_size = vocab_size
        self.eot_token = vocab_size - 1
        self.embed = nn.Embedding(self.vocab_size, self.d_model)
        self.pos_embed = PositionalEmbedding(self.seq_len, self.d_model)
        self.layers = nn.ModuleList(
            [
                TransformerEncoderLayer(self.d_model, self.d_ff, self.n_heads)
                for _ in range(n_layers)
            ]
        )
        self.proj = nn.Linear(self.d_model, self.d_embed)

    def forward(self, x):
        eot_mask = x == self.eot_token
        padding_mask = torch.cumsum(eot_mask, dim=-1) == 0 | eot_mask
        x = self.embed(x)
        x = self.pos_embed(x)
        for layer in self.layers:
            x = layer(x, padding_mask)
        x = x[:, 0]
        x = self.proj(x)
        x = F.normalize(x, dim=-1)
        return x


class ViTImageEncoder(nn.Module):
    def __init__(
        self, d_embed, n_layers, d_model, seq_len, n_heads, input_channels, patch_size
    ):
        super().__init__()
        self.d_embed = d_embed
        self.n_layers = n_layers
        self.d_model = d_model
        self.d_ff = self.d_model * 4
        self.seq_len = seq_len
        self.n_heads = n_heads
        self.input_channels = input_channels
        self.patch_size = patch_size
        self.embed = nn.Conv2d(
            self.input_channels,
            self.d_model,
            self.patch_size,
            self.patch_size,
            bias=False,
        )
        self.class_embed = nn.Parameter(
            torch.randn([self.d_model]) / self.d_model**0.5
        )
        self.pos_embed = PositionalEmbedding(self.seq_len + 1, self.d_model)
        self.layers = nn.ModuleList(
            [
                TransformerEncoderLayer(self.d_model, self.d_ff, self.n_heads)
                for _ in range(n_layers)
            ]
        )
        self.proj = nn.Linear(self.d_model, self.d_embed)

    def forward(self, x):
        x = self.embed(x)
        x = x.reshape([x.shape[0], x.shape[1], -1]).permute([0, 2, 1])
        x = torch.cat(
            [self.class_embed[None, None].repeat([x.shape[0], 1, 1]), x], dim=1
        )
        x = self.pos_embed(x)
        for layer in self.layers:
            x = layer(x)
        x = x[:, 0]
        x = self.proj(x)
        x = F.normalize(x, dim=-1)
        return x


class CLOOBModel(nn.Module):
    def __init__(self, config, image_encoder, text_encoder):
        super().__init__()
        self.config = config
        self.image_encoder = image_encoder
        self.text_encoder = text_encoder
        self.normalize = transforms.Normalize(
            self.config["image_encoder"]["normalize"]["mean"],
            self.config["image_encoder"]["normalize"]["std"],
        )
        assert self.config["text_encoder"]["tokenizer"] == "clip"
        self.tokenize = open_clip.tokenize


def get_pt_model(config):
    assert config["image_encoder"]["type"] == "ViT"
    assert config["text_encoder"]["type"] == "transformer"
    image_encoder = ViTImageEncoder(
        config["d_embed"],
        config["image_encoder"]["n_layers"],
        config["image_encoder"]["d_model"],
        (config["image_encoder"]["image_size"] // config["image_encoder"]["patch_size"])
        ** 2,
        config["image_encoder"]["n_heads"],
        config["image_encoder"]["input_channels"],
        config["image_encoder"]["patch_size"],
    )
    text_encoder = TextEncoder(
        config["d_embed"],
        config["text_encoder"]["n_layers"],
        config["text_encoder"]["d_model"],
        config["text_encoder"]["text_size"],
        config["text_encoder"]["n_heads"],
        config["text_encoder"]["vocab_size"],
    )
    return CLOOBModel(config, image_encoder, text_encoder)


def map_to_tensor(x):
    return {k: {k2: torch.tensor(v2) for k2, v2 in v.items()} for k, v in x.items()}


def convert_jax_vit_image_params(params):
    base = "vi_t_image_encoder"
    pt_base = "image_encoder"
    params = map_to_tensor(params)
    state = {}
    for k, v in params.items():
        names = k.split("/")
        if k == base:
            state[f"{pt_base}.class_embed"] = v["class_embed"]
        elif names[1] == "embed":
            state[f"{pt_base}.embed.weight"] = v["w"].permute([3, 2, 0, 1])
        elif names[1] == "pos_embed":
            state[f"{pt_base}.pos_embed.weight"] = v["w"]
        elif names[1] == "proj":
            state[f"{pt_base}.proj.weight"] = v["w"].T
            state[f"{pt_base}.proj.bias"] = v["b"]
        elif names[1].startswith("layer"):
            layer_num = int(names[1].partition("_")[2])
            if names[2] == "self_attention":
                if names[3] == "layer_norm":
                    state[f"{pt_base}.layers.{layer_num}.attn.norm.weight"] = v["scale"]
                    state[f"{pt_base}.layers.{layer_num}.attn.norm.bias"] = v["offset"]
                elif names[3] == "multi_head_attention":
                    if names[4] == "query":
                        state[f"{pt_base}.layers.{layer_num}.attn.query.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.query.bias"] = v["b"]
                    elif names[4] == "key":
                        state[f"{pt_base}.layers.{layer_num}.attn.key.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.key.bias"] = v["b"]
                    elif names[4] == "value":
                        state[f"{pt_base}.layers.{layer_num}.attn.value.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.value.bias"] = v["b"]
                    elif names[4] == "linear":
                        state[f"{pt_base}.layers.{layer_num}.attn.out.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.out.bias"] = v["b"]
            elif names[2] == "feed_forward":
                if names[3] == "layer_norm":
                    state[f"{pt_base}.layers.{layer_num}.ff.norm.weight"] = v["scale"]
                    state[f"{pt_base}.layers.{layer_num}.ff.norm.bias"] = v["offset"]
                elif names[3] == "linear_0":
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_0.weight"] = v["w"].T
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_0.bias"] = v["b"]
                elif names[3] == "linear_1":
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_1.weight"] = v["w"].T
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_1.bias"] = v["b"]
    return state


def convert_jax_text_params(params):
    base = "text_encoder"
    pt_base = "text_encoder"
    params = map_to_tensor(params)
    state = {}
    for k, v in params.items():
        names = k.split("/")
        if names[1] == "embed":
            state[f"{pt_base}.embed.weight"] = v["embeddings"]
        elif names[1] == "pos_embed":
            state[f"{pt_base}.pos_embed.weight"] = v["w"]
        elif names[1] == "proj":
            state[f"{pt_base}.proj.weight"] = v["w"].T
            state[f"{pt_base}.proj.bias"] = v["b"]
        elif names[1].startswith("layer"):
            layer_num = int(names[1].partition("_")[2])
            if names[2] == "self_attention":
                if names[3] == "layer_norm":
                    state[f"{pt_base}.layers.{layer_num}.attn.norm.weight"] = v["scale"]
                    state[f"{pt_base}.layers.{layer_num}.attn.norm.bias"] = v["offset"]
                elif names[3] == "multi_head_attention":
                    if names[4] == "query":
                        state[f"{pt_base}.layers.{layer_num}.attn.query.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.query.bias"] = v["b"]
                    elif names[4] == "key":
                        state[f"{pt_base}.layers.{layer_num}.attn.key.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.key.bias"] = v["b"]
                    elif names[4] == "value":
                        state[f"{pt_base}.layers.{layer_num}.attn.value.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.value.bias"] = v["b"]
                    elif names[4] == "linear":
                        state[f"{pt_base}.layers.{layer_num}.attn.out.weight"] = v[
                            "w"
                        ].T
                        state[f"{pt_base}.layers.{layer_num}.attn.out.bias"] = v["b"]
            elif names[2] == "feed_forward":
                if names[3] == "layer_norm":
                    state[f"{pt_base}.layers.{layer_num}.ff.norm.weight"] = v["scale"]
                    state[f"{pt_base}.layers.{layer_num}.ff.norm.bias"] = v["offset"]
                elif names[3] == "linear_0":
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_0.weight"] = v["w"].T
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_0.bias"] = v["b"]
                elif names[3] == "linear_1":
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_1.weight"] = v["w"].T
                    state[f"{pt_base}.layers.{layer_num}.ff.linear_1.bias"] = v["b"]
    return state


def get_pt_params(config, checkpoint):
    assert config["image_encoder"]["type"] == "ViT"
    assert config["text_encoder"]["type"] == "transformer"
    cloob_params = pickle.load(open(checkpoint, "rb"))["params"]
    state = {
        **convert_jax_vit_image_params(cloob_params[0]),
        **convert_jax_text_params(cloob_params[1]),
    }
    return state



import torch
from torch import nn
from torch.nn import functional as F


_kernels = {
    "linear": [1 / 8, 3 / 8, 3 / 8, 1 / 8],
    "cubic": [
        -0.01171875,
        -0.03515625,
        0.11328125,
        0.43359375,
        0.43359375,
        0.11328125,
        -0.03515625,
        -0.01171875,
    ],
    "lanczos3": [
        0.003689131001010537,
        0.015056144446134567,
        -0.03399861603975296,
        -0.066637322306633,
        0.13550527393817902,
        0.44638532400131226,
        0.44638532400131226,
        0.13550527393817902,
        -0.066637322306633,
        -0.03399861603975296,
        0.015056144446134567,
        0.003689131001010537,
    ],
}
_kernels["bilinear"] = _kernels["linear"]
_kernels["bicubic"] = _kernels["cubic"]


class Downsample2d(nn.Module):
    def __init__(self, kernel="linear", pad_mode="reflect"):
        super().__init__()
        self.pad_mode = pad_mode
        kernel_1d = torch.tensor([_kernels[kernel]])
        self.pad = kernel_1d.shape[1] // 2 - 1
        self.register_buffer("kernel", kernel_1d.T @ kernel_1d)

    def forward(self, x):
        x = F.pad(x, (self.pad,) * 4, self.pad_mode)
        weight = x.new_zeros(
            [x.shape[1], x.shape[1], self.kernel.shape[0], self.kernel.shape[1]]
        )
        indices = torch.arange(x.shape[1], device=x.device)
        weight[indices, indices] = self.kernel.to(weight)
        return F.conv2d(x, weight, stride=2)


class Upsample2d(nn.Module):
    def __init__(self, kernel="linear", pad_mode="reflect"):
        super().__init__()
        self.pad_mode = pad_mode
        kernel_1d = torch.tensor([_kernels[kernel]]) * 2
        self.pad = kernel_1d.shape[1] // 2 - 1
        self.register_buffer("kernel", kernel_1d.T @ kernel_1d)

    def forward(self, x):
        x = F.pad(x, ((self.pad + 1) // 2,) * 4, self.pad_mode)
        weight = x.new_zeros(
            [x.shape[1], x.shape[1], self.kernel.shape[0], self.kernel.shape[1]]
        )
        indices = torch.arange(x.shape[1], device=x.device)
        weight[indices, indices] = self.kernel.to(weight)
        return F.conv_transpose2d(x, weight, stride=2, padding=self.pad * 2 + 1)



import torch
from torch import nn
import torch.nn.functional as F
import torchvision

from perceptor.transforms.resize import resize
from .interface import LossInterface


class StyleTransfer(LossInterface):
    def __init__(self, style_images=None):
        """Original style transfer loss"""
        super().__init__()
        self.model = torchvision.models.vgg19(pretrained=True).features
        self.model.eval()
        self.model.requires_grad_(False)

        if style_images is not None:
            self.encodings = nn.ParameterList(
                [
                    nn.Parameter(tensor, requires_grad=False)
                    for tensor in self.encode(style_images)
                ]
            )

    def encode(self, images):
        if images.shape[-2:] != (256, 256):
            images = resize(images, out_shape=(256, 256))
        return get_vgg_activations(self.model, [images])

    def loss(self, encodings_a, encodings_b):
        vgg_loss = [
            F.l1_loss(encodings_a[i], encodings_b[i]) for i in range(len(encodings_a))
        ]
        vgg_loss_gram = [
            F.l1_loss(gram_matrix(encodings_a[i]), gram_matrix(encodings_b[i]))
            for i in range(len(encodings_a))
        ]

        vgg_loss = 5 * vgg_loss[2] + 15 * vgg_loss[3] + 2 * vgg_loss[4]
        vgg_loss_gram = (
            5**2 * 5e3 * vgg_loss_gram[2]
            + 15**2 * 5e3 * vgg_loss_gram[3]
            + 2**2 * 5e3 * vgg_loss_gram[4]
        )

        return (vgg_loss + vgg_loss_gram) * 0.001

    def forward(self, images_a, images_b=None):
        if images_b is None:
            encodings_b = self.encodings
        else:
            encodings_b = self.encode(images_b)
        return self.loss(self.encode(images_a), encodings_b)


def gram_matrix(input):
    a, b, c, d = input.size()
    features = input.view(a * b, c * d)
    G = torch.mm(features, features.t())
    return G.div(a * b * c * d)


def get_vgg_activations(vgg16, X_list):
    activations = [(0, 4), (4, 9), (9, 16), (16, 23), (23, 30)]
    for i, (start_index, end_index) in enumerate(activations):
        X_list.append(vgg16[start_index:end_index](X_list[i]))
    return X_list



import torch
import lantern

from perceptor.transforms.clamp_with_grad import clamp_with_grad


class Predictions(lantern.FunctionalBase):
    from_diffused_latents: lantern.Tensor.dims("NCHW")
    from_indices: lantern.Tensor.dims("N")
    predicted_noise: lantern.Tensor.dims("NCHW")
    schedule_alphas: lantern.Tensor
    schedule_sigmas: lantern.Tensor

    @property
    def device(self):
        return self.predicted_noise.device

    def indices(self, indices):
        if isinstance(indices, float) or isinstance(indices, int):
            indices = torch.as_tensor(indices)
        if indices.ndim == 0:
            indices = indices[None]
        if indices.ndim != 1:
            raise ValueError("indices must be a scalar or a 1D tensor")
        return indices.long().to(self.device)

    def alphas(self, indices):
        return self.schedule_alphas[self.indices(indices)][:, None, None, None].to(
            self.device
        )

    def sigmas(self, indices):
        return self.schedule_sigmas[self.indices(indices)][:, None, None, None].to(
            self.device
        )

    @property
    def from_alphas(self):
        return self.alphas(self.from_indices)

    @property
    def from_sigmas(self):
        return self.sigmas(self.from_indices)

    @property
    def denoised_latents(self):
        return (
            self.from_diffused_latents - self.from_sigmas * self.predicted_noise
        ) / self.from_alphas.clamp(min=1e-7)

    def step(self, to_indices, eta=0.0):
        """
        Reduce noise level to `to_indices`

        Args:
            to_indices: Union[Tensor, Tensor.shape("N"), float]
            eta: float

        Returns:
            diffused_images: torch.Tensor.shape("NCHW")
        """
        to_alphas, to_sigmas = self.alphas(to_indices), self.sigmas(to_indices)

        if eta > 0.0:
            # If eta > 0, adjust the scaling factor for the predicted noise
            # downward according to the amount of additional noise to add
            ddim_sigma = (
                eta
                * (to_sigmas**2 / self.from_sigmas**2).sqrt()
                * (1 - self.from_alphas**2 / to_alphas**2).sqrt()
            )
            adjusted_sigma = (to_sigmas**2 - ddim_sigma**2).sqrt()

            # Recombine the predicted noise and predicted denoised image in the
            # correct proportions for the next step
            to_diffused_latents = (
                self.denoised_latents * to_alphas
                + self.predicted_noise * adjusted_sigma
            )

            # Add the correct amount of fresh noise
            noise = torch.randn_like(to_diffused_latents)
            to_diffused_latents += noise * ddim_sigma
        else:
            to_diffused_latents = (
                self.denoised_latents * to_alphas + self.predicted_noise * to_sigmas
            )

        return to_diffused_latents

        # TODO: do not need to calculate denoised latents? this could introduce errors?

    def correction(self, previous: "Predictions") -> "Predictions":
        # k-diffusion has alphas=1 always so this should not work
        # corrected_diffused_xs = (
        #     previous.from_diffused_xs
        #     + (self.from_sigmas - previous.from_sigmas) * (self.eps - previous.eps) / 2
        # )
        # return diffusion_space.decode(corrected_diffused_xs)

        # blurry
        # return Predictions(
        #     from_diffused_images=previous.from_diffused_images,
        #     from_indices=previous.from_indices,
        #     velocities=(self.velocities + previous.velocities) / 2,
        # )

        # looks ok but no apparent difference from double budget
        return previous.forced_denoised(
            (self.denoised_latents + previous.denoised_latents) / 2
        )

    def reverse_step(self, to_indices):
        if (torch.as_tensor(self.from_indices) > torch.as_tensor(to_indices)).any():
            raise ValueError("from_indices must be less than to_indices")

        # replaced_noise_sigma = (
        #     self.sigmas(to_t) ** 2 - self.sigmas(from_t) ** 2
        # ).sqrt()
        # to_eps = (
        #     from_eps * self.sigmas(from_t)
        #     + torch.randn_like(from_eps) * replaced_noise_sigma
        # ) / self.sigmas(to_t)
        # to_eps = eps
        # to_diffused = self.diffuse(from_denoised, to_t, noise=to_eps)
        # return to_diffused

        to_alphas, to_sigmas = self.alphas(to_indices), self.sigmas(to_indices)
        return self.denoised_latents * to_alphas + self.predicted_noise * to_sigmas
        # TODO: do not need to calculate denoied latents? this could introduce errors?

    def resample(self, resample_indices):
        """
        Harmonizing resampling from https://github.com/andreas128/RePaint
        """
        return (
            self.denoised_latents * self.from_alphas
            + self.resample_noise(resample_indices) * self.from_sigmas
        )

    def resample_noise(self, resample_indices):
        if (
            torch.as_tensor(self.from_indices) < torch.as_tensor(resample_indices)
        ).any():
            raise ValueError("from_indices must be greater than resample_indices")
        resampled_noise_sigma = (
            self.sigmas(resample_indices) * self.predicted_noise
            + (
                self.from_sigmas**2 - self.sigmas(resample_indices) ** 2
            ).sqrt() * torch.randn_like(self.predicted_noise)
        )  # fmt: skip
        return resampled_noise_sigma / self.from_sigmas

    def noisy_reverse_step(self, to_indices):
        to_alphas, to_sigmas = self.alphas(to_indices), self.sigmas(to_indices)

        noise_sigma = self.from_sigmas * self.predicted_noise + (
            to_sigmas**2 - self.from_sigmas**2
        ).sqrt() * torch.randn_like(self.predicted_noise)

        return self.denoised_latents * to_alphas + noise_sigma

    def guided(self, guiding, guidance_scale=0.5, clamp_value=1e-6) -> "Predictions":
        return self.replace(
            predicted_noise=self.predicted_noise
            + guidance_scale
            * self.from_sigmas
            * guiding.clamp(-clamp_value, clamp_value)
            / clamp_value
        )

    def dynamic_threshold(self, quantile=0.95) -> "Predictions":
        """
        Thresholding heuristic from imagen paper
        """
        dynamic_threshold = torch.quantile(
            self.denoised_latents.flatten(start_dim=1).abs(), quantile, dim=1
        ).clamp(min=1.0)
        denoised_latents = (
            clamp_with_grad(
                self.denoised_latents,
                -dynamic_threshold,
                dynamic_threshold,
            )
            # / dynamic_threshold
            # imagen's dynamic thresholding divides by threshold but this makes the images gray
        )
        return self.forced_denoised_latents(denoised_latents)

    def forced_denoised_latents(self, denoised_latents) -> "Predictions":
        # if (self.from_sigmas >= 1e-3).all():
        #     predicted_noise = (
        #         self.from_diffused_latents - denoised_latents * self.from_alphas
        #     ) / self.from_sigmas
        # else:
        #     predicted_noise = self.predicted_noise
        # return self.replace(
        #     velocities=self.from_alphas * predicted_noise
        #     - self.from_sigmas * denoised_latents
        # )
        pass

    def forced_predicted_noise(self, predicted_noise) -> "Predictions":
        return self.replace(predicted_noise=predicted_noise)

    def wasserstein_distance(self):
        sorted_noise = self.predicted_noise.flatten(start_dim=1).sort(dim=1)[0]
        n = sorted_noise.shape[1]
        margin = 0.5 / n
        points = torch.linspace(margin, 1 - margin, sorted_noise.shape[1])
        expected_noise = torch.distributions.Normal(0, 1).icdf(points)
        return (sorted_noise - expected_noise[None].to(sorted_noise)).abs().mean()

    def wasserstein_square_distance(self):
        sorted_noise = self.predicted_noise.flatten(start_dim=1).sort(dim=1)[0]
        n = sorted_noise.shape[1]
        margin = 0.5 / n
        points = torch.linspace(margin, 1 - margin, sorted_noise.shape[1])
        expected_noise = torch.distributions.Normal(0, 1).icdf(points)
        return (sorted_noise - expected_noise[None].to(sorted_noise)).square().mean()

    def classifier_free_guidance(
        self, positive_predictions: "Predictions", guidance_scale=7.0
    ) -> "Predictions":
        return self.replace(
            predicted_noise=self.predicted_noise
            + (positive_predictions.predicted_noise - self.predicted_noise)
            * guidance_scale
        )



import os
from huggingface_hub import hf_hub_url, cached_download

from .model import CLIP
from .processor import RuCLIPProcessor

MODELS = {
    "ruclip-vit-base-patch32-224": dict(
        repo_id="sberbank-ai/ruclip-vit-base-patch32-224",
        filenames=["bpe.model", "config.json", "pytorch_model.bin"],
    ),
    "ruclip-vit-base-patch16-224": dict(
        repo_id="sberbank-ai/ruclip-vit-base-patch16-224",
        filenames=["bpe.model", "config.json", "pytorch_model.bin"],
    ),
    "ruclip-vit-large-patch14-224": dict(
        repo_id="sberbank-ai/ruclip-vit-large-patch14-224",
        filenames=["bpe.model", "config.json", "pytorch_model.bin"],
    ),
    "ruclip-vit-large-patch14-336": dict(
        repo_id="sberbank-ai/ruclip-vit-large-patch14-336",
        filenames=["bpe.model", "config.json", "pytorch_model.bin"],
    ),
    "ruclip-vit-base-patch32-384": dict(
        repo_id="sberbank-ai/ruclip-vit-base-patch32-384",
        filenames=["bpe.model", "config.json", "pytorch_model.bin"],
    ),
    "ruclip-vit-base-patch16-384": dict(
        repo_id="sberbank-ai/ruclip-vit-base-patch16-384",
        filenames=["bpe.model", "config.json", "pytorch_model.bin"],
    ),
}


def load(name, device="cpu", cache_dir="models", use_auth_token=None):
    """Load a ruCLIP model
    Parameters
    ----------
    name : str
        A model name listed in ruclip.MODELS.keys()
    device : Union[str, torch.device]
        The device to put the loaded model
    cache_dir: str
        path to download the model files; by default, it uses "models"
    Returns
    -------
    clip : torch.nn.Module
        The ruCLIP model
    clip_processor : ruclip.processor.RuCLIPProcessor
        A ruCLIP processor which performs tokenization and image preprocessing
    """
    assert name in MODELS, f"All models: {MODELS.keys()}"
    config = MODELS[name]
    repo_id = config["repo_id"]
    cache_dir = os.path.join(cache_dir, name)
    for filename in config["filenames"]:
        config_file_url = hf_hub_url(repo_id=repo_id, filename=f"{filename}")
        cached_download(
            config_file_url,
            cache_dir=cache_dir,
            force_filename=filename,
            use_auth_token=use_auth_token,
        )

    clip = CLIP.from_pretrained(cache_dir).eval().to(device)
    clip_processor = RuCLIPProcessor.from_pretrained(cache_dir)
    return clip, clip_processor
